import re
import math
import warnings
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from typing import Any
from typing import Self
from typing import List
from typing import Dict
from typing import Tuple
from typing import Literal
from typing import overload
from typing import Sequence
from typing import Generator
from typing import Callable
from numpy.typing import NDArray
from matplotlib.text import Text
from matplotlib.axes import Axes
from matplotlib.lines import Line2D
from matplotlib.figure import Figure
from matplotlib.legend import Legend
from matplotlib.artist import Artist
from matplotlib.ticker import Formatter
from matplotlib.ticker import FuncFormatter
from matplotlib.ticker import StrMethodFormatter
from matplotlib.typing import HashableList
from matplotlib.patches import Patch

from .plotter import Stripe
from .plotter import StripeLine
from .plotter import StripeSpan

from .._typing import MosaicLayout
from .._typing import NumericSample1D
from .._typing import ShareAxisProperty
from .._typing import LegendHandlesLabels

from ..strings import STR

from ..constants import KW
from ..constants import LABEL
from ..constants import DEFAULT

from ..statistics import SpecLimits
from ..statistics import ProcessEstimator


__all__ = [
    "flat_unique",
    "LabelFacets",
    "AxesFacets",
    "StripesFacets",]


def flat_unique(nested: NDArray | List[List]) -> List:
    """Flatten the given array and return unique elements while
    preserving the order."""
    if isinstance(nested, list):
        nested = np.array(nested)
    return list(pd.Series(nested.flatten()).unique())


class AxesFacets:
    """A class for creating a grid of subplots with customizable sharing
    and sizing options.    
    
    The class provides flexible handling the created Axes instances 
    through its flat property and iteration capabilities:

    - **Flat property:** 
      Accesses Axes in a flattened view from left-to-right, 
      top-to-bottom
    - **Axes property:** 
      Maintains the 2D structure for grid-based operations
    - **Iteration:** 
      Yields each axis instance coming from the Flat property 
      sequentially to simplify plotting
    The class supports two different ways to create subplot layouts:

    Parameters
    ----------
    nrows : int, optional
        Number of rows of subplots in the grid, by default 1.
    ncols : int, optional
        Number of columns of subplots in the grid, by default 1.
    sharex : bool or {'none', 'all', 'row', 'col'}, optional
        Controls sharing of properties along the x-axis,
        by default 'none'.
    sharey : bool or {'none', 'all', 'row', 'col'}, optional
        Controls sharing of properties along the y-axis,
        by default 'none'.
    width_ratios : array-like of length ncols, optional
        Relative widths of the columns, by default None.
    height_ratios : array-like of length nrows, optional
        Relative heights of the rows, by default None.
    stretch_figsize : bool | float | Tuple[float, float], optional
        If True, the height and width of the figure are stretched based 
        on the number rows and columns in the axes grid. If a float is 
        provided, the figure size is stretched by the given factor. If a 
        tuple of two floats is provided, the figure size is stretched by 
        the given factors for the x and y axis, respectively.
        by default False.
    **kwds : dict, optional
        Additional keyword arguments to pass to the function 
        `plt.subplots` or to the function `plt.subplot_mosaic` if a 
        mosaic was specified.

    Examples
    --------
    1. Using nrows and ncols:
    Creates a regular grid of subplots with specified dimensions
    ```python
    # Creates a 2x3 grid of regular subplots
    facets = AxesFacets(nrows=2, ncols=3)
    ```

    2. Using mosaic:
    Creates a layout with custom subplot arrangements and spanning
    ```python
    import daspi as dsp
    # Creates a layout with spanning cells
    layout = (
        'AAB',
        'CDB')
    facets = dsp.AxesFacets(mosaic=layout)
    ```
    
    3. Create a seaborn-style jointplot layout with custom size ratios:
    ```python
    import daspi as dsp
    # Creates a layout with main scatter plot and marginal distributions
    layout = [
        ['hist_x', '.'],      # '.' creates an empty/blank Axes
        ['scatter', 'hist_y']]
    facets = AxesFacets(
        mosaic=layout,
        width_ratios=[4, 1],   # Make the marginal y-hist narrower
        height_ratios=[1, 4]   # Make the marginal x-hist shorter
        )
    ```

    This creates a figure with:
    - A main scatter plot in the bottom-left
    - A marginal histogram on top for x-distribution
    - A marginal histogram on right for y-distribution
    - Top-right cell is automatically empty using the '.' notation
    - Proportional spacing using width and height ratios

    The '.' character is a special notation in matplotlib's mosaic layout 
    that automatically creates an empty/invisible Axes, which is more 
    efficient than creating a visible Axes and then hiding it.
    """

    figsize: Tuple[float, float]
    """The figsize passed when creating the subplots."""
    figure: Figure
    """The figure instance to label."""
    axes: NDArray
    """A 2D array containing the Axes instances of the figure."""
    _nrows: int
    """Number of Axes rows in the grid"""
    _ncols: int
    """Number of Axes columns in the grid"""
    _sharex: ShareAxisProperty
    """Controls sharing of properties along the x-axis."""
    _sharey: ShareAxisProperty
    """Controls sharing of properties along the y-axis."""
    _ax: Axes | None
    """The current axes being worked on"""
    _default: Axes | None
    """The default Axes instance after iterating over this class. It is 
    the first Axes object in the grid if there is only one, otherwise 
    None."""
    mosaic: HashableList | None
    """A visual layout of how the Axes are arranged labeled as strings."""
    mosaic_pattern = re.compile(r'[\s+]?(\S+)[\s+]?')

    def __init__(
            self,
            *,
            nrows: int | None = None,
            ncols: int | None = None,
            mosaic: MosaicLayout | None = None,
            sharex: ShareAxisProperty = 'none', 
            sharey: ShareAxisProperty = 'none', 
            width_ratios: Sequence[float] | None = None,
            height_ratios: Sequence[float] | None = None, 
            stretch_figsize: bool | float | Tuple[float, float] = False,
            **kwds
            ) -> None:
        assert not all(arg is not None for arg in (nrows, ncols, mosaic)), (
            'Either nrows and ncols or mosaic must be provided, but not both.')
        
        if isinstance(mosaic, str):
            mosaic = self.mosaic_pattern.findall(mosaic)
        if isinstance(mosaic, (list, tuple)):
            self.mosaic = [[c for c in r] for r in mosaic]
            self._nrows = len(self.mosaic)
            self._ncols = len(self.mosaic[0])
            assert all(self._ncols == len(row) for row in self.mosaic), (
                f'Received a non-rectangular grid for the {mosaic=}.')
            self._sharex = True if sharex in (True, 'all') else False
            self._sharey = True if sharey in (True, 'all') else False
        else:
            self.mosaic = None
            self._sharex = sharex
            self._sharey = sharey
            self._nrows = nrows if isinstance(nrows, int) else 1
            self._ncols = ncols if isinstance(ncols, int) else 1

        figsize = kwds.pop('figsize', plt.rcParams['figure.figsize'])
        
        if stretch_figsize is False:
            stretch_x = stretch_y = 1
        elif stretch_figsize is True:
            stretch_x = (1 + math.log(self._ncols, math.e))
            stretch_y = (1 + math.log(self._nrows, math.e))
        elif isinstance(stretch_figsize, (tuple, list)):
            stretch_x, stretch_y = stretch_figsize
        elif isinstance(stretch_figsize, (int, float)):
            stretch_x = stretch_y = stretch_figsize
        else:
            raise TypeError(f'{stretch_figsize=} is not supported.')
        self.figsize = (
            (stretch_x * figsize[0], stretch_y * figsize[1]))

        _kwds: Dict[str, Any] = dict(
            sharex=self._sharex,
            sharey=self._sharey, 
            figsize=self.figsize,
            width_ratios=width_ratios,
            height_ratios=height_ratios,
            layout='tight',
            ) | kwds
        if self.mosaic:
            self.figure, axes = plt.subplot_mosaic(mosaic=self.mosaic, **_kwds)
            self.axes = np.array(
                [[axes.get(key, None) for key in row] for row in self.mosaic])
        else:
            self.figure, self.axes = plt.subplots(
                nrows=self._nrows, ncols=self._ncols, squeeze=False, **_kwds)

        self._default = self.axes[0, 0] if self.shape == (1, 1) else None
        self._ax = self._default

    @property
    def ax(self) -> Axes | None:
        """Get the axes that is currently being worked on. This property
        is automatically kept current when iterating through this 
        class (read-only)."""
        return self._ax
    
    @property
    def nrows(self) -> int:
        """Get the number of Axes rows in the grid (read-only)."""
        return self._nrows
    
    @property
    def ncols(self) -> int:
        """Get the number of Axes columns in the grid (read-only)."""
        return self._ncols
    
    @property
    def shape(self) -> Tuple[int, ...]:
        """Get the shape of the axes array (read-only)."""
        return self.axes.shape

    @property
    def sharex(self) -> ShareAxisProperty:
        """Get the sharing of properties along the x-axis (read-only)."""
        return self._sharex

    @property
    def sharey(self) -> ShareAxisProperty:
        """Get the sharing of properties along the y-axis (read-only)."""
        return self._sharey
    
    @property
    def flat(self) -> List[Axes]:
        """Get a list of all the Axes in the grid.

        Returns
        -------
        List[Axes]
            A list of all the Axes in the grid.
        """
        return [ax for ax in flat_unique(self.axes) if isinstance(ax, Axes)]
    
    def __iter__(self) -> Generator[Axes, Self, None]:
        """Iterate over the axes in the grid.

        Returns
        -------
        Generator[Axes, Self, None]
            The generator object that yields the axes in the grid.
        """
        def ax_gen(self) -> Generator[Axes, Self, None]:
            for ax in self.flat:
                self._ax = ax
                yield ax
            self._ax = self._default
        return ax_gen(self)
    
    def __next__(self) -> Axes:
        """Get the next axes in the grid.

        Returns
        -------
        Axes
            The next axes in the grid."""
        return next(self)
    
    @overload
    def __getitem__(self, index: Tuple[slice, slice]) -> NDArray:...

    @overload
    def __getitem__(self, index: Tuple[int, slice]) -> NDArray:...

    @overload
    def __getitem__(self, index: Tuple[slice, int]) -> NDArray:...
    
    @overload
    def __getitem__(self, index: int | Tuple[int, int]) -> Axes:...
    
    def __getitem__(self, index: int | Tuple[int | slice, int | slice]) -> Any:
        """Get the axes at the specified index in the grid. If the index
        is a number, the axes are fetched from left to right and from 
        top to bottom. If two numbers are specified, the axis is fetched 
        directly in the grid and it is then also possible to fetch an 
        axis across multiple indices if the axis is spanned across 
        multiple columns or rows.

        Parameters
        ----------
        index : int | Tuple[int, int]
            The index of the axes in the grid.

        Returns
        -------
        Axes
            The axes at the given index in the grid.
        """
        if isinstance(index, tuple):
            return self.axes[index]
        return self.flat[index]

    def __len__(self) -> int:
        """Get the total number of axes in the grid.

        Returns
        -------
        int
            The total number of axes in the grid.
        """
        return len(self.flat)


class StripesFacets:
    """A class for creating location and spread width lines, 
    specification limits and/or confidence interval areas as stripes on 
    each Axes. The location and spread (and their confidence bands) 
    represent the data per axes.

    Parameters
    ----------
    target : ArrayLike
        The target data for the estimation.
    target_on_y : bool
        Whether the target data is on the y-axis or not.
    single_axes : bool
        Whether a single axis is used for the chart or whether the 
        stripes are used on multiple axes. This affects how the 
        predefined legend labels are handled. If True, the corresponding
        position values are added to the label.
    stripes : List[Type[Stripe]], optional
        Additional non-predefined stripes to be added to the chart, 
        by default [].
    mean : bool, optional
        Whether to include a mean line on the chart, by default False.
    median : bool, optional
        Whether to include a median line on the chart, by default False.
    control_limits : bool, optional
        Whether to include control limits on the chart, by default False.
    spec_limits : SpecLimits, optional
        The specification limits for the chart, by default SpecLimits().
    confidence : float, optional
        The confidence level for the confidence intervals,
        by default None.
    strategy : {'eval', 'fit', 'norm', 'data'}, optional
        Which strategy should be used to determine the control limits
        (process spread):
        - eval: The strategy is determined according to the given 
          evaluate function. If none is given, the internal evaluate
          method is used.
        - fit: First, the distribution that best represents the process
          data is searched for and then the agreed process spread is 
          calculated
        - norm: it is assumed that the data is subject to normal
          distribution. The variation tolerance is then calculated as 
          agreement * standard deviation
        - data: The quantiles for the process variation tolerance are
          read directly from the data. by default 'norm'
    
    agreement : float or int, optional
        Specify the tolerated process variation for which the control 
        limits are to be calculated.
        If int, the spread is determined using the normal distribution 
        agreementsigma, e.g. agreement = 6 -> 6sigma ~ covers 99.75 % of 
        the data. The upper and lower permissible quantiles are then 
        calculated from this.
        If float, the value must be between 0 and 1.This value is then 
        interpreted as the acceptable proportion for the spread, e.g. 
        0.9973 (which corresponds to ~ 6 sigma) by default 6
    **kwds : dict, optional
        Additional keyword arguments to pass to the ProcessEstimator.
    """

    __slots__ = (
        'estimation', '_confidence', 'single_axes', 'stripes', 'target_on_y')
    
    estimation: ProcessEstimator
    """The process estimator object."""
    _confidence: float | None
    """The confidence level for the confidence intervals."""
    single_axes: bool
    """Whether a single axis is used for the chart or whether the 
    stripes are used on multiple axes. This affects how the predefined
    legend labels are handled. If True, the corresponding position
    values are added to the label."""
    stripes: Dict[str, Stripe]
    """The stripes to plot on the chart."""
    target_on_y: bool
    """Whether the target is on the y-axis or the x-axis."""
    
    def __init__(
        self,
        target: NumericSample1D,
        *,
        target_on_y: bool,
        single_axes: bool,
        stripes: List[Stripe] = [], 
        mean: bool = False,
        median: bool = False,
        control_limits: bool = False,
        spec_limits: SpecLimits = SpecLimits(),
        confidence: float | None = None,
        strategy: Literal['eval', 'fit', 'norm', 'data'] = 'norm',
        agreement: float | int = 6,
        **kwds) -> None:
        self.single_axes = single_axes
        self._confidence = confidence
        self.estimation = ProcessEstimator(
            samples=target,
            spec_limits=spec_limits,
            strategy=strategy,
            agreement=agreement,
            **kwds)
        self.target_on_y = target_on_y
        self.stripes = {s.identity: s for s in stripes}
        add_confidence_span = confidence is not None
        
        self.add_specification_limit_stripes()
        if mean:
            self.add_mean_stripes(add_confidence_span)
        if median:
            self.add_median_stripes(add_confidence_span)
        if control_limits:
            self.add_control_limit_stripes(add_confidence_span)
        
    @property
    def confidence(self) -> float:
        """Get the confidence level for the confidence intervals
        (read-only)."""
        if self._confidence is None:
            return DEFAULT.CONFIDENCE_LEVEL
        return self._confidence
    
    @property
    def orientation(self) -> Literal['horizontal', 'vertical']:
        """The orientation of the chart."""
        return 'horizontal' if self.target_on_y else 'vertical'
    
    @property
    def _kwds(self) -> Dict[str, Any]:
        kwds = dict(
            orientation=self.orientation,
            show_position=self.single_axes)
        return kwds

    @property
    def _confidence_label(self) -> str:
        """The lable used for confidence stripes."""
        if self._confidence is not None:
            label = f'{100*self._confidence:.0f} \\%-{STR["ci"]}'
        else:
            label = ''
        return label
    
    def add_mean_stripes(self, add_confidence_span: bool) -> None:
        """Add mean stripes to the plot.

        This method adds a stripe representing the mean of the target 
        data, using the `StripeLine` class. If `add_confidence_span` is 
        `True`, additional stripes representing the confidence interval 
        for the mean will also be added using the `StripeSpan` class.

        Parameters
        ----------
        add_confidence_span : bool
            Whether to add confidence stripes for the mean.
        """
        line = StripeLine(
            label=r'\bar x',
            position=self.estimation.mean,
            **(self._kwds | KW.MEAN_LINE))
        self.stripes[line.identity] = line

        if add_confidence_span:
            low, upp = self.estimation.mean_ci(self.confidence)
            span = StripeSpan(
                label=self._confidence_label,
                lower_position=low,
                upper_position=upp,
                color=line.color,
                **(self._kwds | KW.STRIPES_CONFIDENCE))
            self.stripes[f'{span.identity}_mean'] = span

    def add_median_stripes(self, add_confidence_span: bool) -> None:
        """Add median stripes to the plot.

        This method adds a stripe representing the median of the target 
        data, using the `StripeLine` class. If `add_confidence_span` is 
        `True`, additional stripes representing the confidence interval 
        for the median will also be added using the `StripeSpan` class.

        Parameters
        ----------
        add_confidence_span : bool
            Whether to add confidence stripes for the median.
        """
        line = StripeLine(
            label=r'\tilde x',
            position=self.estimation.median,
            **(self._kwds | KW.MEDIAN_LINE))
        self.stripes[line.identity] = line

        if add_confidence_span:
            low, upp = self.estimation.median_ci(self.confidence)
            span = StripeSpan(
                label=self._confidence_label,
                lower_position=low,
                upper_position=upp,
                color=line.color,
                **(self._kwds | KW.STRIPES_CONFIDENCE))
            self.stripes[f'{span.identity}_median'] = span
    
    def add_control_limit_stripes(self, add_confidence_span: bool) -> None:
        """Add control limit stripes to the plot.
        
        This method adds stripes representing the lower and upper 
        control limits for the target data.
        
        The stripes are added using the `StripeLine` class, with the 
        label and position set based on the control limits from the 
        `estimation` attribute. If `add_confidence_span` is `True`,
        additional stripes representing the confidence interval for the
        control limits will also be added using the `StripeSpan` class.
        
        Parameters
        ----------
        add_confidence_span : bool
            Whether to add confidence stripes for the control limits.
        """
                
        kw_line = self._kwds | KW.CONTROL_LINE
        line_low = StripeLine(
            label=r'x_{' + f'{self.estimation.q_low:.4f}' + '}',
            position=self.estimation.lcl,
            **kw_line)
        line_upp = StripeLine(
            label=r'x_{' + f'{self.estimation.q_upp:.4f}' + '}',
            position=self.estimation.ucl,
            **kw_line)
        self.stripes[line_low.identity] = line_low
        self.stripes[line_upp.identity] = line_upp
        
        if add_confidence_span:
            kw_span = self._kwds | KW.STRIPES_CONFIDENCE
            low, upp = self.estimation.stdev_ci(self.confidence)
            span_low = StripeSpan(
                label=self._confidence_label,
                position=self.estimation.lcl,
                width=upp - low,
                color=line_low.color,
                **kw_span)
            span_upp = StripeSpan(
                label=self._confidence_label,
                position=self.estimation.ucl,
                width=upp - low,
                color=line_upp.color,
                **kw_span)
            self.stripes[f'{span_low.identity}_low'] = span_low
            self.stripes[f'{span_upp.identity}_upp'] = span_upp
    
    def add_specification_limit_stripes(self) -> None:
        """Add specification limit stripes to the plot.
        
        This method adds stripes representing the lower and upper 
        specification limits to the plot. If a specification limit is
        `None`, it will be skipped.
        
        The stripes are added using the `StripeLine` class, with the 
        label and position set based on the specification limits from
        the `estimation` attribute.
        """
        kwds = self._kwds | KW.SPECIFICATION_LINE
        _labels = (STR['lsl'], STR['usl'])
        _limits = self.estimation.spec_limits.to_tuple()
        for label, limit in zip(_labels, _limits):
            if float('-inf') < limit < float('inf'):
                line = StripeLine(label=str(label), position=limit, **kwds)
                self.stripes[line.identity] = line
    
    def handles_labels(self) -> LegendHandlesLabels:
        """Get the legend handles and labels for the plot.
        
        This method processes the `self.stripes` dictionary to extract
        the legend handles and labels for the plot. It separates the 
        handles and labels for line stripes and span stripes, and 
        returns them as a tuple of handles and a tuple of labels.
        
        Returns
        -------
        LegendHandlesLabels:
            A tuple containing the legend handles and labels.
        """
        handles_labels_lines = {}
        handles_labels_spans = {}
        processed = set()
        for stripe in self.stripes.values():
            if stripe.identity in processed:
                continue
            if isinstance(stripe, StripeLine):
                handles_labels_lines[stripe.label] = stripe.handle
            else:
                handles_labels_spans[stripe.label] = stripe.handle
            processed.add(stripe.identity)
        handles_labels = handles_labels_lines | handles_labels_spans
        return tuple(handles_labels.values()), tuple(handles_labels.keys())

    def draw(self, ax: Axes) -> None:
        """Draw the stripes on the specified Axes.

        Parameters
        ----------
        ax : Axes
            The Axes on which to draw the stripes.
        """
        for stripe in self.stripes.values():
            stripe(ax)


class LabelFacets:
    """
    A class for adding labels and titles to facets of a figure with 
    advanced formatting.

    This class provides comprehensive labeling and formatting
    capabilities for matplotlib figures, including support for custom
    formatters, label  rotation, alignment control, and automatic margin
    adjustment for optimal layout.

    Parameters
    ----------
    axes : AxesFacets
        A AxesFacets instance containing the subplots' Axes and their 
        arrangement.
    fig_title : str, optional
        Main title that should be displayed at the top of the chart,
        by default ''.
    sub_title : str, optional
        Subtitle, which should appear directly below the main title and
        slightly smaller than it, by default ''.
    xlabel, ylabel: str or Tuple[str, ...], optional
        The axis label(s) of the figure. To label multiple axes with 
        different names, provide a tuple; otherwise, provide a string,
        by default ''.
    xlabel_formatter, ylabel_formatter : Formatter | Callable | str | None, optional
        Advanced formatters for axis tick labels with multiple input 
        types:
        
        - **String format templates**: Simple format strings using 
          Python's string formatting syntax (e.g., '{:.2f}', '{:.1e}',
          '${:.0f}', '{:.1%}') for quick and intuitive formatting
        - **Callable functions**: Custom functions that take one
          argument (value) or two arguments (value, position) and return 
          formatted strings for complete control
        - **Matplotlib Formatters**: Any matplotlib.ticker.Formatter
          instance for advanced formatting scenarios
        - **None**: Use matplotlib's default formatting
        
        The formatter automatically handles matplotlib compatibility by
        wrapping simple callables in FuncFormatter and converting string
        templates to appropriate formatter types. By default, None.
        
    xlabel_angle, ylabel_angle : float, optional
        Rotation angle for x and y axis tick labels in degrees. Positive
        values rotate counter-clockwise, negative values rotate
        clockwise. The chart automatically adjusts margins to
        accommodate rotated labels and prevent clipping. Common values: 
        0 (horizontal), 45 (diagonal), 90 (vertical). By default, 0.
        
    xlabel_align, ylabel_align : str, optional
        Alignment for x and y axis tick labels relative to their tick 
        marks:
        
        - **For x-axis**: 'left', 'center', 'right' control horizontal
          alignment of the label text relative to the tick position
        - **For y-axis**: 'bottom', 'center', 'top' control vertical
          alignment, where the string values map to vertical alignment
          for better API consistency
          
        Alignment is particularly useful with rotated labels to achieve
        optimal positioning and readability. By default, 'center'.
    
    info : bool or str, optional
        Indicates whether to include an info text at the lower left 
        corner in the figure. If True, the date and user are
        automatically  added. If a string is provided, it's appended to
        the automatic date/user information, separated by a comma.
        By default, False.
    rows: Tuple[str, ...], optional
        The row labels of the figure for faceted plots, by default ().
    cols: Tuple[str, ...], optional
        The column labels of the figure for faceted plots,
        by default ().
    row_title : str, optional
        The title of the rows for faceted plots, by default ''.
    col_title : str, optional
        The title of the columns for faceted plots, by default ''.
    axes_titles : Tuple[str, ...]
        Title for each Axes, useful for JointCharts with multiple
        subplots, by default ().
    legend_data : Dict[str, LegendHandlesLabels], optional
        The legends to be added to the figure. The key is used as the 
        legend title, and the values must be a tuple of tuples, where
        the inner tuple contains a handle as a Patch or Line2D artist
        and a label as a string, by default {}.

    Examples
    --------
    Basic usage with string formatters:
    
    ```python
    axes = AxesFacets(nrows=1, ncols=1)
    label_facets = LabelFacets(
        axes=axes,
        fig_title='Temperature Analysis',
        xlabel='Time (hours)',
        ylabel='Temperature',
        ylabel_formatter='{:.1f}°C',  # String template
        xlabel_formatter='{:.0f}h'    # String template
    )
    label_facets.draw()
    ```
    
    Custom callable formatters:
    
    ```python
    def scientific_formatter(value):
        if abs(value) >= 1000:
            return f'{value:.1e}'
        return f'{value:.2f}'
    
    label_facets = LabelFacets(
        axes=axes,
        ylabel_formatter=scientific_formatter,
        xlabel_formatter=lambda x: f'${x:,.0f}'  # Lambda formatter
    )
    ```
    
    Rotation and alignment for readability:
    
    ```python
    label_facets = LabelFacets(
        axes=axes,
        xlabel='Product Categories',
        xlabel_angle=45,      # Diagonal rotation
        xlabel_align='right', # Right-align for better rotation appearance
        ylabel_formatter='{:.1%}'  # Percentage format
    )
    ```
    
    Advanced faceted plot labeling:
    
    ```python
    label_facets = LabelFacets(
        axes=multi_axes,
        fig_title='Sales Analysis by Region',
        sub_title='Q3 2024 Performance',
        row_title='Geographic Region',
        col_title='Product Category',
        rows=('North', 'South', 'East', 'West'),
        cols=('Electronics', 'Clothing', 'Books'),
        xlabel_formatter='${:,.0f}',
        ylabel_formatter='{:.1f}%',
        info='Source: Sales database'
    )
    ```

    Notes
    -----
    - String formatters are automatically converted to 
      matplotlib-compatible formatters using intelligent type detection.
    - Margins are automatically adjusted when using rotated labels to
      prevent clipping and ensure optimal layout.
    - The class handles both single axes and complex multi-axes layouts.
    - All formatting options work consistently across different
      matplotlib backends and figure configurations.
    - Legend positioning is automatically calculated to avoid
      overlapping with other chart elements.
    """

    axes: AxesFacets
    """A AxesFacets instance containing the subplots' Axes and their 
    arrangement."""
    figure: Figure
    """The figure instance to label."""
    fig_title: str
    """The title to display at the top of the chart."""
    sub_title: str
    """The subtitle to display directly below the title of the chart."""
    xlabel: str | Tuple[str, ...]
    """The x-axis label(s) of the figure."""
    ylabel: str | Tuple[str, ...]
    """The y-axis label(s) of the figure."""
    xlabel_formatter: Callable | Formatter | None
    """Function to format the x-axis tick labels."""
    ylabel_formatter: Callable | Formatter | None
    """Function to format the y-axis tick labels."""
    xlabel_angle: float
    """Rotation angle for x-axis tick labels in degrees."""
    ylabel_angle: float
    """Rotation angle for y-axis tick labels in degrees."""
    xlabel_align: Literal['left', 'center', 'right']
    """Alignment for x-axis tick labels."""
    ylabel_align: Literal['bottom', 'center', 'top']
    """Alignment for y-axis tick labels."""
    info: bool | str
    """Indicates whether to include an info text in the figure."""
    rows: Tuple[str, ...]
    """The row labels of the figure."""
    cols: Tuple[str, ...]
    """The column labels of the figure."""
    row_title: str
    """The title of the rows."""
    col_title: str
    """The title of the columns."""
    axes_titles: Tuple[str, ...]
    """The titles of each axes."""
    legend_data: Dict[str, LegendHandlesLabels]
    """The legend_data to be added to the figure."""
    _legend: Legend | None
    """Figure legend if one is added"""
    _size: Tuple[int, int]
    """The size of the figure in pixels."""
    _margin: Dict[str, int]
    """The required margin for the labels of the image in pixels."""
    labels: Dict[str, Text]
    """The labels that were added to the image as text objects."""

    def __init__(
            self,
            axes: AxesFacets,
            *,
            fig_title: str = '', 
            sub_title: str = '',
            xlabel: str | Tuple[str, ...] = '',
            ylabel: str | Tuple[str, ...] = '',
            xlabel_formatter: Callable | None = None,
            ylabel_formatter: Callable | None = None,
            xlabel_angle: float = 0,
            ylabel_angle: float = 0,
            xlabel_align: Literal['left', 'center', 'right'] = 'center',
            ylabel_align: Literal['bottom', 'center', 'top'] = 'center',
            info: bool | str = False,
            rows: Tuple[str, ...] = (),
            cols: Tuple[str, ...] = (),
            row_title: str = '',
            col_title: str = '',
            axes_titles: Tuple[str, ...] = (),
            legend_data: Dict[str, LegendHandlesLabels] = {}
            ) -> None:
        self.axes = axes
        self.figure = self.axes.figure
        self.fig_title = fig_title
        self.sub_title = sub_title
        self.xlabel = xlabel
        self.ylabel = ylabel
        self.xlabel_formatter = self._prepare_formatter(xlabel_formatter)
        self.ylabel_formatter = self._prepare_formatter(ylabel_formatter)
        self.xlabel_angle = xlabel_angle
        self.ylabel_angle = ylabel_angle
        self.xlabel_align = xlabel_align
        self.ylabel_align = ylabel_align
        self.rows = rows
        self.cols = cols
        self.row_title = row_title
        self.col_title = col_title
        self.axes_titles = axes_titles
        self.info = info
        self.legend_data = legend_data
        self._legend = None
        self._size = (
            int(self.figure.get_figwidth() * self.figure.dpi),
            int(self.figure.get_figheight() * self.figure.dpi))
        self._margin = dict(left=0, bottom=0, right=0, top=0)
        self.labels = {}
    
    def _prepare_formatter(
            self,
            formatter: Formatter | Callable | str | None
            ) -> Formatter | None:
        """Prepare a formatter for use with matplotlib.
        
        If the formatter is a callable that takes only one argument,
        wrap it in a FuncFormatter. If it's already a Formatter or
        takes two arguments, use it as-is. If it's a string, convert it
        to a FuncFormatter for robust formatting.
        
        Parameters
        ----------
        formatter : Formatter | Callable | str | None
            The formatter to prepare. String formatters can use either
            positional format (e.g., '{:.2e}') or keyword format 
            (e.g., '{x:.2e}').
            
        Returns
        -------
        Formatter | None
            A matplotlib-compatible formatter or None
        """
        if formatter is None:
            return None
        
        if isinstance(formatter, Formatter):
            return formatter
        
        if isinstance(formatter, str):
            # Convert string formatters to FuncFormatter for robust handling
            # This handles both positional '{:.2e}' and keyword '{x:.2e}' formats
            if '{x' in formatter.lower() or '{pos' in formatter.lower():
                # Keyword format - use StrMethodFormatter
                return StrMethodFormatter(formatter)
            else:
                # Positional format - use FuncFormatter with string formatting
                return FuncFormatter(lambda x, pos, fmt=formatter: fmt.format(x))

        if callable(formatter):
            # Check if the callable accepts only one argument
            import inspect
            sig = inspect.signature(formatter)
            if len(sig.parameters) == 1:
                # Wrap single-argument callable in FuncFormatter
                return FuncFormatter(lambda x, pos, f=formatter: f(x))
            else:
                # Assume it's already a two-argument callable
                return FuncFormatter(formatter)
        
        return None
    
    @property
    def x_aligned(self) -> float:
        """Get aligned x position as fraction of figure width. This is
        used for title, subtitle and info (read-only)."""
        return KW._margin + LABEL.X_ALIGNEMENT / self._size[0]

    @property
    def margin_rectangle(self) -> Tuple[float, float, float, float]:
        """Get rectangle of margins around the subplots used for the
        additional labels as fraction of figure size (read-only)."""
        margins = (
            self._margin['left'] / self._size[0],
            self._margin['bottom'] / self._size[1],
            max(1 - (self._margin['right'] / self._size[0]), 0.5),
            max(1 - (self._margin['top'] / self._size[1]), 0.5))
        return margins
    
    @property
    def legend_fraction(self) -> float:
        """Get fraction of figure size for legend (read-only)."""
        return self.legend_width / self.figure.get_figwidth()

    @staticmethod
    def estimate_height(text: Text) -> int:
        """Get the estimated size of the text in the figure in pixels."""
        dpi = text.figure.dpi if text.figure else plt.rcParams['figure.dpi']
        return int(int(text.get_fontsize()) * LABEL.PPI * dpi) + LABEL.PADDING
    
    @staticmethod
    def estimate_rotation_margin(angle_degrees: float, base_margin: int = 25) -> int:
        """Estimate additional margin needed for rotated tick labels.
        
        Parameters
        ----------
        angle_degrees : float
            Rotation angle in degrees
        base_margin : int, optional
            Base margin for calculating rotated label space, by default 25
            
        Returns
        -------
        int
            Additional margin in pixels needed for the rotation
        """
        if abs(angle_degrees) < 5:  # No significant rotation
            return 0
        
        angle_rad = abs(angle_degrees) * np.pi / 180
        # Use sine to estimate the additional perpendicular space needed
        return int(base_margin * np.sin(angle_rad))
    
    @staticmethod
    def get_legend_artists(legend: Legend) -> List[Artist]:
        """Get the inner children of a legend.

        Parameters
        ----------
        legend : Legend
            The legend object.

        Returns
        -------
        List[Artist]:
            The artists representing the inner children of the legend.
        """
        return legend.get_children()[0].get_children()

    @property
    def legend(self) -> Legend | None:
        """Get legend added to figure (read-only)."""
        return self._legend
    
    @property
    def legend_artists(self) -> List[Artist]:
        """Get legend artists (read-only)."""
        if self.legend is None:
            return []
        return self.get_legend_artists(self.legend)
    
    @property
    def legend_width(self) -> int:
        """Get width of legend in pixels (read-only)."""
        if self.legend is None:
            return 0
        return int(self.legend.get_window_extent().width)

    def _add_legend(
            self, handles: Tuple[Patch |Line2D, ...], labels: Tuple[str, ...],
            title: str) -> None:
        """Adds a legend at the right side of the figure. If there is 
        already one, the existing one is extended with the new one
        
        Parameters
        ----------
        handles: Tuple[Patch | Line2D, ...]
            A list of Artists (lines, patches) to be added to the
            legend.
        labels : Tuple[str, ...]
            The labels must be in the same order as the corresponding 
            plots were drawn. If no labels are given, the handles and 
            labels of the first axes are used.
        title : str
            Title for the given handles and labels. 
        """
        legend = Legend(
            self.figure, handles, labels, title=title, **KW.LEGEND)
        
        if not self.legend:
            self.figure.legends.append(legend) # type: ignore
            self._legend = legend
        else:
            new_artists = self.get_legend_artists(legend)
            self.legend_artists.extend(new_artists)
    
    @staticmethod
    def not_shared(axis) -> bool:
        """Check if an axis is not shared."""
        return len(axis._get_shared_axis()) == 1
    
    def _add_axes_titles(self) -> None:
        """Add the axes titles to the figure. If cols are set, the
        axes titles are not drawn."""
        if not self.axes_titles:
            return
        
        if self.cols:
            warnings.warn('Drawing axes_titles ignored, because cols are set.')
            return

        for ax, title in zip(self.axes.flat, self.axes_titles):
            ax.set_title(title)

    def _apply_axis_formatting(self) -> None:
        """Apply formatters and rotation to all axes before calculating margins.
        This ensures that margin calculations account for the space needed by 
        rotated or formatted labels."""
        for ax in self.axes.flat:
            # Apply formatters
            if self.xlabel_formatter is not None:
                ax.xaxis.set_major_formatter(self.xlabel_formatter)
            if self.ylabel_formatter is not None:
                ax.yaxis.set_major_formatter(self.ylabel_formatter)
            
            # Apply rotation and alignment
            if self.xlabel_angle != 0 or self.xlabel_align != 'center':
                ax.tick_params(axis='x', rotation=self.xlabel_angle)
                # Apply horizontal alignment for x-axis labels
                for label in ax.get_xticklabels():
                    label.set_horizontalalignment(self.xlabel_align)
                    
            if self.ylabel_angle != 0 or self.ylabel_align != 'center':
                ax.tick_params(axis='y', rotation=self.ylabel_angle)
                # Apply vertical alignment for y-axis labels
                for label in ax.get_yticklabels():
                    label.set_verticalalignment(self.ylabel_align)

    def _add_left_labels(self) -> None:
        """Add the ylabel or ylabs. If only one is present, a label will
        be added centrally for all axes. Otherwise, a label will be added
        for all axes that do not have a shared axes."""
        self._margin['left'] = 0

        if self.ylabel:
            if isinstance(self.ylabel, str):
                _kwds = KW.YLABEL
                _kwds['x'] += self._margin['left'] / self._size[0]
                _text = self.figure.text(s=self.ylabel, **_kwds)
                self._margin['left'] += self.estimate_height(_text)
                self.labels['ylabel'] = _text
            else:
                for ax, ylabel in zip(self.axes.flat, self.ylabel):
                    if self.not_shared(ax.yaxis) or ax in self.axes[:, 0]:
                        ax.set(ylabel=ylabel)
        
        # Add extra margin for rotated y-axis labels
        if self.ylabel_angle != 0:
            extra_margin = self.estimate_rotation_margin(self.ylabel_angle, base_margin=20)
            self._margin['left'] += extra_margin

    def _add_bottom_labels(self) -> None:
        """Insert the xlabel or xlabs. If only one is present, a label 
        will be added centrally for all axes. Otherwise, a label will be 
        added for all axes that do not have a shared axes.
        Also insert an info text in the bottom left-hand corner of the 
        figure. By default, the info text contains today's date and the 
        user name. If attribute `info` is a string, it is added to the 
        info text separated by a comma."""
        self._margin['bottom'] = 0

        if self.info:
            _kwds = KW.INFO | dict(x=self.x_aligned)
            info_text = f'{STR.TODAY} {STR.USERNAME}'
            if isinstance(self.info, str):
                info_text = f'{info_text}, {self.info}'
            _text = self.figure.text(s=info_text, **_kwds)
            self._margin['bottom'] += self.estimate_height(_text)
            self.labels['info'] = _text
        
        if self.xlabel:
            if isinstance(self.xlabel, str):
                _kwds = KW.XLABEL
                _kwds['y'] += self._margin['bottom'] / self._size[1]
                _text = self.figure.text(s=self.xlabel, **_kwds)
                self._margin['bottom'] += self.estimate_height(_text)
                self.labels['xlabel'] = _text
            else:
                for ax, xlabel in zip(self.axes.flat, self.xlabel):
                    if self.not_shared(ax.xaxis) or ax in self.axes[-1, :]: 
                        ax.set(xlabel=xlabel)
        
        # Add extra margin for rotated x-axis labels
        if self.xlabel_angle != 0:
            extra_margin = self.estimate_rotation_margin(self.xlabel_angle, base_margin=30)
            self._margin['bottom'] += extra_margin

    def _add_right_labels(self) -> None:
        """Add row labels and row title to the figure. The title is 
        added to figure object, the rows to the axes objects."""
        self._margin['right'] = self.legend_width
        if self.legend_width > 0:
            self._margin['right'] += 2 * LABEL.PADDING

        if self.row_title:
            _kwds = KW.ROW_TITLE
            _kwds['x'] -= self._margin['right'] / self._size[0]
            _text = self.figure.text(s=self.row_title, **_kwds)
            self._margin['right'] += self.estimate_height(_text)
            self.labels['row_title'] = _text

        if self.rows:
            for ax, label in zip(self.axes[:, -1], self.rows):
                _kwds = KW.ROW_LABEL | {'transform': ax.transAxes}
                _text = ax.text(s=label, **_kwds)
                self.labels[f'row_{label}'] = _text

    def _add_top_labels(self) -> None:
        """Add all provided labels at the top of the charts. These 
        labels include the figure title, sub title, column title and 
        column labels."""
        self._margin['top'] = 0

        if self.fig_title:
            _kwds = KW.FIG_TITLE | dict(x=self.x_aligned)
            _text = self.figure.text(s=self.fig_title, **_kwds)
            self._margin['top'] += self.estimate_height(_text) + LABEL.PADDING
            self.labels['fig_title'] = _text

        if self.sub_title:
            _kwds = KW.SUB_TITLE | dict(x=self.x_aligned)
            _kwds['y'] -= self._margin['top'] / self._size[1]
            _text = self.figure.text(s=self.sub_title, **_kwds)
            self._margin['top'] += self.estimate_height(_text) + LABEL.PADDING
            self.labels['sub_title'] = _text
        
        if self.col_title:
            _kwds = KW.COL_TITLE
            _kwds['y'] -= self._margin['top'] / self._size[1]
            _text = self.figure.text(s=self.col_title, **_kwds)
            self._margin['top'] += self.estimate_height(_text)
            self.labels['col_title'] = _text

        if self.cols:
            for ax, label in zip(self.axes[0, :], self.cols):
                _kwds = KW.COL_LABEL | {'transform': ax.transAxes}
                _text = ax.text(s=label, **_kwds)
                self.labels[f'col_{label}'] = _text
    
    def _remove_last_labelpad(self) -> None:
        """Remove the padding added for last added labels."""
        for pos, margin in self._margin.items():
            if margin > LABEL.PADDING:
                self._margin[pos] = margin - LABEL.PADDING
    
    def _adjust_centered_labels(self) -> None:
        """Adjust the position of all centered labels according to the 
        current margin settings."""
        lr_adjustment = (
            (self._margin['left'] - self._margin['right']) / self._size[0] / 2)
        bt_adjustment = (
            (self._margin['bottom'] - self._margin['top']) / self._size[1] / 2)
        if 'xlabel' in self.labels:
            self.labels['xlabel'].set_x(KW.XLABEL['x'] + lr_adjustment)
        if 'ylabel' in self.labels:
            self.labels['ylabel'].set_y(KW.YLABEL['y'] + bt_adjustment)
        if 'col_title' in self.labels:
            self.labels['col_title'].set_x(KW.COL_TITLE['x'] + lr_adjustment)
        if 'row_title' in self.labels:
            self.labels['row_title'].set_y(KW.ROW_TITLE['y'] + bt_adjustment)
        if self.legend:
            bbox = self.legend.get_bbox_to_anchor()
            self.legend.set_bbox_to_anchor((
                bbox.x1 / self._size[0],
                (bbox.y1 - self._margin['top']) / self._size[1]))
    
    # TODO: this does not work as expected
    def clear(self) -> None:
        """Remove all the label facets."""
        for label in self.labels.values():
            label.remove()
            if label in self.figure.texts:
                self.figure.texts.remove(label)
            del(label)
        self.labels = {}

        if self.legend:
            self.legend.remove()
            if self.legend in self.figure.legends:
                self.figure.legends.remove(self.legend)
            self._legend = None
        
        self._margin = dict(left=0, bottom=0, right=0, top=0)
    
    def draw(self) -> None:
        """Draw all the label facets to the figure."""
        self.clear()

        # Apply axis formatting first, before calculating margins
        self._apply_axis_formatting()
        
        self._add_axes_titles()
        for title, (handles, labels) in self.legend_data.items():
            self._add_legend(handles, labels, title)
        self._add_left_labels()
        self._add_bottom_labels()
        self._add_right_labels()
        self._add_top_labels()

        self._remove_last_labelpad()
        self._adjust_centered_labels()
            
        self.figure.tight_layout(rect=self.margin_rectangle)



import warnings
import pandas as pd

from typing import Any
from typing import List
from typing import Self
from typing import Dict
from typing import Type
from typing import Tuple
from typing import Literal
from typing import Generator
from matplotlib.axes import Axes
from matplotlib.ticker import PercentFormatter
from pandas.core.frame import DataFrame
from scipy.stats._distn_infrastructure import rv_continuous

from .chart import JointChart
from .chart import check_label_order

from .plotter import Bar
from .plotter import Line
from .plotter import Stem
from .plotter import Pareto
from .plotter import Plotter
from .plotter import Scatter
from .plotter import MeanTest
from .plotter import Beeswarm
from .plotter import StripeLine
from .plotter import StripeSpan
from .plotter import SkipSubplot
from .plotter import HideSubplot
from .plotter import Probability
from .plotter import GaussianKDE
from .plotter import BlandAltman
from .plotter import QuantileBoxes
from .plotter import CenterLocation
from .plotter import GaussianKDEContour
from .plotter import ParallelCoordinate
from .plotter import LinearRegressionLine
from .plotter import CategoricalObservation
from .plotter import CapabilityConfidenceInterval

from ..anova import LinearModel
from ..anova import GageRnRModel
from ..anova import GageStudyModel

from ..strings import STR

from ..constants import DIST
from ..constants import COLOR
from ..constants import ANOVA
from ..constants import DEFAULT
from ..constants import CATEGORY

from ..statistics import SpecLimits
from ..statistics import ProcessEstimator


__all__ = [
    'ResidualsCharts',
    'ParameterRelevanceCharts',
    'PairComparisonCharts',
    'BivariateUnivariateCharts',
    'ProcessCapabilityAnalysisCharts',
    'GageStudyCharts',
    'GageRnRCharts',
    'PairwiseMatrixCharts',]


class ParameterRelevanceCharts(JointChart):
    """
    Provides a set of charts for visualizing the relevance of a linear
    regression model's parameters.

    The `ParameterRelevanceCharts` class takes a `LinearModel` instance 
    and generates a set of two charts for visualizing the relevance
    of the model's parameters:
    - Pareto chart of the parameter standardized effects
    - Pareto chart of the Sum of Squares for each parameter

    On the first chart, the standardized effect of each parameter is 
    visualized as a Pareto chart. The red line indicates the alpha
    level of significance.

    Parameters
    ----------
    linear_model : LinearModel
        The linear regression model whose parameters will be visualized.
    drop_intercept : bool, optional
        Whether to drop the intercept from the model, by default True.
    stretch_figsize : bool | float | Tuple[float, float], optional
        If True, the height and width of the figure are stretched based 
        on the number rows and columns in the axes grid. If a float is 
        provided, the figure size is stretched by the given factor. If a 
        tuple of two floats is provided, the figure size is stretched by 
        the given factors for the x and y axis, respectively.
        by default False.
    
    Examples
    --------

    ```python
    import daspi as dsp
    import pandas as pd

    df = dsp.load_dataset('painkillers-dissolution')
    model = dsp.LinearModel(
        source=df,
        target='dissolution',
        features=['employee', 'stirrer', 'brand', 'catalyst', 'water'],
        covariates=['temperature', 'preparation'],
        order=2)
    df_gof = pd.concat(model.recursive_elimination())
    dsp.ParameterRelevanceCharts(model).plot().stripes().label(info=True)
    ```

    ![ParameterRelevanceChartsDocstringExample](https://raw.githubusercontent.com/j4ggr/DaSPi/main/docs/img/precast_parameter-relevance-charts_example.png)
    """
    __slots__ = ('lm')

    lm: LinearModel
    """The linear regression model whose parameters are visualized."""

    def __init__(
            self,
            linear_model: LinearModel,
            drop_intercept: bool = True,
            stretch_figsize: bool | float | Tuple[float, float] = False
            ) -> None:
        self.lm = linear_model
        effects =  self.lm.effects()
        if ANOVA.INTERCEPT in effects and drop_intercept:
            effects = effects.drop(ANOVA.INTERCEPT)
        data = (pd
            .concat([self.lm.anova('I'), effects], axis=1)
            .reset_index(drop=False)
            .rename(columns={'index': ANOVA.SOURCE}))

        super().__init__(
            source=data,
            target=(ANOVA.EFFECTS, ANOVA.TABLE_COLNAMES[1]),
            feature=ANOVA.SOURCE,
            target_on_y=False,
            ncols=2,
            nrows=1,
            stretch_figsize=stretch_figsize)
        
    def plot(self) -> Self: # type: ignore
        """Generates a set of two charts for visualizing the relevance 
        of the model's parameters:
        - Pareto chart of the parameter standardized effects
        - Pareto chart of the Sum of Squares for each parameter
        
        Returns
        -------
        Self: 
            The `ParameterRelevanceCharts` instance, for method chaining.

        Notes
        -----
        The `plot()` method will generate two charts. The first chart
        will contain the Pareto chart of the parameter standardized
        effects. The second chart will contain the Pareto chart of the
        Sum of Squares for each parameter.

        The red line in the Pareto charts represents the threshold
        for the effect of a parameter. The threshold is calculated as
        the alpha risk where the parameter is not relevant.
        """
        super().plot(
            Pareto, no_percentage_line=True, skip_na='all')
        super().plot(
            Pareto, highlight=ANOVA.RESIDUAL, skip_na='all')
        return self
    
    def stripes(self) -> Self: # type: ignore
        """Adds a line at position 0 for each subplot except for the 
        probability plot. This line represents the fit of the model."""
        threshold = StripeLine(
            label=STR['charts_label_alpha_th'].format(alpha=self.lm.alpha),
            position=self.lm.effect_threshold,
            orientation='vertical',
            color=COLOR.SPECIAL_LINE)
        for chart in self.itercharts():
            if chart == self.charts[0]:
                chart.stripes([threshold])
                break
        return self

    def label(self, info: bool | str = False, **kwds) -> Self: # type: ignore
        """Adds titles and labels to the charts generated by the 
        `plot()` method.
        
        Parameters
        ----------
        info : bool | str, optional
            If `True`, the method will add an informative subtitle to 
            the chart. If a string is provided, it will be used as the 
            subtitle, by default False.
        **kwds
            Additional keyword arguments to be passed to the `label()`
            method of the `JointChart` instance.
        
        Returns
        -------
        Self
            The `ParameterRelevanceCharts` instance, for method chaining.
        """      
        labels = dict(
                fig_title=STR['paramcharts_fig_title'],
                sub_title=STR['paramcharts_sub_title'],
                target_label=(STR['effects_label'], STR['ss_label']),
                feature_label=(STR['paramcharts_feature_label'], ''),
                info = info
            ) | kwds
        super().label(**labels) # type: ignore
        return self


class ResidualsCharts(JointChart):
    """
    Provides a set of charts for visualizing the residuals of a linear 
    regression model.

    The `ResidualsCharts` class takes a `LinearModel` instance and 
    generates a set of four charts:
    - Probability plot of the residuals
    - Gaussian kernel density estimate of the residuals
    - Scatter plot of the predicted values vs. the observed values
    - Line plot of the predicted values vs. the observed values

    The `plot()` method generates the charts, and the `label()` method 
    adds titles and labels to the charts.

    Parameters
    ----------
    linear_model : LinearModel
        The linear regression model whose residuals will be visualized.
    stretch_figsize : bool | float | Tuple[float, float], optional
        If True, the height and width of the figure are stretched based 
        on the number rows and columns in the axes grid. If a float is 
        provided, the figure size is stretched by the given factor. If a 
        tuple of two floats is provided, the figure size is stretched by 
        the given factors for the x and y axis, respectively.
        by default False.
    
    Examples
    --------

    ```python
    import daspi as dsp
    import pandas as pd

    df = dsp.load_dataset('painkillers-dissolution')
    model = dsp.LinearModel(
        source=df,
        target='dissolution',
        features=['employee', 'stirrer', 'brand', 'catalyst', 'water'],
        covariates=['temperature', 'preparation'],
        order=2)
    df_gof = pd.concat(model.recursive_elimination())
    dsp.ResidualsCharts(model).plot().stripes().label(info=True)
    ```

    ![ResidualsChartsDocstringExample](https://raw.githubusercontent.com/j4ggr/DaSPi/main/docs/img/precast_residuals-charts_example.png)
    """
    __slots__ = ('lm')

    lm: LinearModel
    """The linear regression model whose residuals are visualized."""

    def __init__(
            self,
            linear_model: LinearModel,
            stretch_figsize: bool | float | Tuple[float, float] = False
            ) -> None:
        self.lm = linear_model
        super().__init__(
            source=self.lm.residual_data(),
            target=ANOVA.RESIDUAL,
            feature=('', '', ANOVA.PREDICTION, ANOVA.OBSERVATION),
            nrows=2,
            ncols=2,
            sharey=True,
            stretch_figsize=stretch_figsize)
        
    def plot(self) -> Self: # type: ignore
        """Generates a set of four charts for visualizing the residuals 
        of a linear regression model:
        - Probability plot of the residuals
        - Gaussian kernel density estimate of the residuals
        - Scatter plot of the predicted values vs. the observed values
        - Line plot of the predicted values vs. the observed values
        
        Returns
        -------
        Self: 
            The `ResidualsCharts` instance, for method chaining.
        """
        super().plot(Probability, show_fit_ci=True)
        super().plot(GaussianKDE)
        super().plot(Scatter)
        super().plot(Line, kw_call={'marker': 'o'})
        return self
    
    def stripes(self) -> Self: # type: ignore
        """Adds a line at position 0 for each subplot except for the 
        probability plot. This line represents the fit of the model."""
        fit = StripeLine(
            label=STR['fit'],
            position=0,
            orientation='horizontal',
            color=COLOR.SPECIAL_LINE)
        for chart in self.itercharts():
            if chart == self.charts[0]:
                continue
            chart.stripes([fit])
        return self

    def label(self, info: bool | str = False, **kwds) -> Self: # type: ignore
        """Adds titles and labels to the charts generated by the 
        `plot()` method.
        
        Parameters
        ----------
        info : bool | str, optional
            If `True`, the method will add an informative subtitle to 
            the chart. If a string is provided, it will be used as the 
            subtitle, by default False.
        **kwds
            Additional keyword arguments to be passed to the `label()`
            method of the `JointChart` instance.
        
        Returns
        -------
        Self
            The `ResidualsCharts` instance, for method chaining.
        """
        sub_title = f'{self.lm.target} ~ {" + ".join(self.lm.effects().index)}'
        labels = dict(
                fig_title=STR['residcharts_fig_title'],
                sub_title=sub_title,
                target_label=STR['resid_name'],
                feature_label=(
                    STR['charts_flabel_quantiles'],
                    STR['charts_flabel_density'],
                     f'{STR["charts_flabel_predicted"]} {self.lm.target}',
                    STR['charts_flabel_observed']),
                info = info
            ) | kwds
        super().label(**labels) # type: ignore
        return self


class PairComparisonCharts(JointChart):
    """Provides a set of charts for visualizing the pairwise 
    comparison of two variables.
    
    Parameters
    ----------
    source : DataFrame
        The source data.
    target : str or Tuple[str]
        The target variable(s).
    feature : str or Tuple[str]
        The feature variable(s).
    identity : str
        Column name containing identities of each sample, must occur 
        once for each measurement.
    stretch_figsize : bool | float | Tuple[float, float], optional
        If True, the height and width of the figure are stretched based 
        on the number rows and columns in the axes grid. If a float is 
        provided, the figure size is stretched by the given factor. If a 
        tuple of two floats is provided, the figure size is stretched by 
        the given factors for the x and y axis, respectively.
        by default False.
    
    Examples
    --------

    ```python
    import daspi as dsp

    df = dsp.load_dataset('shoe-sole')
    chart = dsp.PairComparisonCharts(
            source=df,
            target='wear',
            feature='status',
            identity='tester'
        ).plot(
        ).label(
            info=True
        )
    ```

    ![PairComparisonChartsDocstringExample](https://raw.githubusercontent.com/j4ggr/DaSPi/main/docs/img/precast_pair-comparison-charts_example.png)
    """

    __slots__ = ('identity')

    identity: str
    """Column name containing identities of each sample."""

    def __init__(
            self,
            source: DataFrame,
            target: str,
            feature: str,
            identity: str,
            stretch_figsize: bool | float | Tuple[float, float] = False
            ) -> None:
        self.identity = identity
        super().__init__(
            source=source,
            target=target,
            feature=feature,
            nrows=1,
            ncols=2,
            width_ratios=[4, 1],
            categorical_feature = (False, True),
            stretch_figsize=stretch_figsize)
    
    def plot(self) -> Self: # type: ignore
        """Generates a set of two charts for visualizing the difference
        between each pair.
        - BlandAltman where the difference is shown on the y-axis
        - ParallelCoordinate including a MeanTest and QuantileBoxes
        plot, to reflect the difference within the absolute values.
        
        Returns
        -------
        Self: 
            The `PairComparisonCharts` instance, for method chaining.
        """
        super().plot(BlandAltman, identity=self.identity, feature_axis='data')
        super().plot(
            ParallelCoordinate, identity=self.identity, show_scatter=False)
        super().plot(MeanTest, n_groups=2, on_last_axes=True)
        super().plot(QuantileBoxes, strategy='fit', on_last_axes=True)
        return self

    def label(self, info: bool | str = False, **kwds) -> Self: # type: ignore
        """Adds titles and labels to the charts generated by the 
        `plot()` method.
        
        Parameters
        ----------
        info : bool | str, optional
            If `True`, the method will add an informative subtitle to 
            the chart. If a string is provided, it will be used as the 
            subtitle, by default False.
        **kwds
            Additional keyword arguments to be passed to the `label()`
            method of the `JointChart` instance.
        
        Returns
        -------
        Self
            The `PairComparisonCharts` instance, for method chaining.
        """
        labels = dict(
                fig_title=STR['paircharts_fig_title'],
                sub_title=STR['paircharts_sub_title'],
                target_label=(True, True),
                feature_label=(True, True),
                info = info
            ) | kwds
        super().label(**labels) # type: ignore
        return self


class BivariateUnivariateCharts(JointChart):
    """Provides a set of charts for visualizing the relationship between
    a target variable and a feature variable.

    This class prepares a grid for drawing a bivariate plot with 
    marginal univariate plots:
    - Bottom left there is a large square axis for the bivariate
      diagram. For example, use a `LinearRegressionLine` plotter to show 
      the dependence of the target variable on the feature variable.
    - Top left there is a small square axis for the univariate
      plot of the feature variable.
    - Bottom right there is a small square axis for the univariate
      plot of the target variable.
    - Top right there is a small square axis hidden as soon a plot is 
      done.

    Parameters
    ----------
    source : DataFrame
        The source data.
    target : str
        The target (dependant) variable.
    feature : str
        The feature (independant) variable.
    hue : str, optional
        The hue variable.
    dodge_univariates : bool, optional
        Whether to dodge the univariate plots, by default False.
    categorical_feature_univariates : bool, optional
        Whether to treat the feature variable for univariate charts as
        categorical, by default False.
    ratios : List[float], optional
        The ratios of the bivariate axes height to the univariate axes 
        height, by default [4, 1].
    stretch_figsize : bool | float | Tuple[float, float], optional
        If True, the height and width of the figure are stretched based 
        on the number rows and columns in the axes grid. If a float is 
        provided, the figure size is stretched by the given factor. If a 
        tuple of two floats is provided, the figure size is stretched by 
        the given factors for the x and y axis, respectively.
        by default False.
    colors: Tuple[str, ...], optional
        Tuple of unique colors used for hue categories as hex or str. If
        not provided, the default colors will be used, by default ().
    
    Examples
    --------

    ```python
    import daspi as dsp

    df = dsp.load_dataset('painkillers-dissolution')
    hue = 'brand'
    n_groups = df.groupby(hue).ngroups
    chart = dsp.BivariateUnivariateCharts(
            source=df,
            target='dissolution',
            feature='temperature',
            hue=hue,
            dodge_univariates=True,
        ).plot_univariates(
            dsp.MeanTest, n_groups=n_groups
        ).plot_univariates(
            dsp.QuantileBoxes, strategy='data'
        ).plot_bivariate(
            dsp.LinearRegressionLine, show_fit_ci=True
        ).label(
            fig_title='Regression and distribution analysis',
            sub_title='Painkillers dissolution time vs. temperature',
            feature_label='Water temperature (°C)',
            target_label='Dissolution time (s)',
            axes_titles=('95 % Bonferroni confidence interval of mean', '', '', ''),
            info=True
        )
    ```
    
    ![BivariateUnivariateChartsDocstringExample](https://raw.githubusercontent.com/j4ggr/DaSPi/main/docs/img/precast_bivariate-univariate-charts_example.png)

    Information about which distribution is used to calculate the
    quantile boxes can be obtained as follows:

    ```python
    brands = (brand for brand in df[hue].unique().tolist()*2)
    for plot in chart.plots:
        if isinstance(plot, dsp.QuantileBoxes):
            print(f'{next(brands)} {plot.target}: {plot.estimation.dist.dist_name}')
    ```
    """
    __slots__ = ('_top_right_hidden')

    _top_right_hidden: bool
    """Whether the top right axis is hidden."""

    def __init__(
            self,
            source: DataFrame,
            target: str,
            feature: str,
            hue: str = '',
            dodge_univariates: bool = False,
            categorical_feature_univariates: bool = False,
            ratios: List[float] = [3, 1],
            stretch_figsize: bool | float | Tuple[float, float] = False,
            colors: Tuple[str, ...] = ()
            ) -> None:
        assert len(ratios) == 2, ('ratios must be a list of two floats')
        self._top_right_hidden = False

        super().__init__(
            source=source,
            target=(
                feature, '',
                target, target),
            feature=(
                '', '',
                feature, ''),
            hue=hue,
            dodge=(
                dodge_univariates, False,
                False, dodge_univariates),
            categorical_feature=(
                categorical_feature_univariates, False,
                False, categorical_feature_univariates),
            target_on_y=(
                False, True,
                True, True),
            nrows=2,
            ncols=2,
            sharex='col',
            sharey='row',
            width_ratios=ratios,
            height_ratios=ratios[::-1],
            stretch_figsize=stretch_figsize,
            colors=colors or CATEGORY.PALETTE)
    
    @property
    def hidden_ax(self) -> Axes:
        """Get the top right axis (read-only)."""
        return self.axes[0, 1]
    
    @property
    def univariate_axs(self) -> List[Axes]:
        """Get the univariate axes (read-only)."""
        return [self.axes[0], self.axes[-1]]
    
    @property
    def bivariate_ax(self) -> Axes:
        """Get the bivariate axis (read-only)."""
        return self.axes[1, 0]
    
    def hide_top_right_ax(self) -> None:
        """Hides the top right axis if it is not already hidden."""
        if self._top_right_hidden:
            return

        HideSubplot(self.hidden_ax)()
        self._top_right_hidden = True
    
    def plot( # type: ignore
            self,
            plotter: Type[Plotter],
            kind: Literal['univariate', 'bivariate'],
            *,
            kw_call: Dict[str, Any] = {},
            kw_where: Dict[str, Any] = {},
            **kwds
            ) -> Self:
        if kind == 'bivariate':
            self.plot_bivariate(
                plotter,
                kw_call=kw_call,
                kw_where=kw_where,
                **kwds)
        elif kind == 'univariate':
            self.plot_univariates(
                plotter,
                kw_call=kw_call,
                kw_where=kw_where,
                **kwds)
        else:
            raise ValueError('kind must be either "bivariate" or "univariate"')
        return self
    
    def plot_univariates(
            self,
            plotter: Type[Plotter],
            *,
            kw_call: Dict[str, Any] = {},
            kw_where: Dict[str, Any] = {},
            **kwds
            ) -> Self:
        """Plots the univariate plots on the top left and bottom right
        axes.

        Parameters
        ----------
        plotter : Type[Plotter]
            The plotter to use for the univariate plots.
        kw_call : Dict[str, Any]
            Additional keyword arguments for the plotter call method.
        kw_where : Dict[str, Any]
            Additional keyword arguments for the where method used to
            filter the data.
        **kwds
            Additional keyword arguments to be passed to the super plot
            method.

        Returns
        -------
        Self
            The `BivariateUnivariateCharts` instance, for method 
            chaining.
        """
        self.hide_top_right_ax()
        for ax in self.axes:
            if ax in self.univariate_axs:
                super().plot(
                    plotter,
                    kw_call=kw_call,
                    kw_where=kw_where,
                    on_last_axes=False,
                    **kwds)
            else:
                super().plot(SkipSubplot)
        return self

    def plot_bivariate(
            self,
            plotter: Type[Plotter],
            *,
            kw_call: Dict[str, Any] = {},
            kw_where: Dict[str, Any] = {},
            **kwds) -> Self:
        """Plots the bivariate plot on the bottom left axis.

        Parameters
        ----------
        plotter : Type[Plotter]
            The plotter to use for the univariate plots.
        kw_call : Dict[str, Any]
            Additional keyword arguments for the plotter call method.
        kw_where : Dict[str, Any]
            Additional keyword arguments for the where method used to
            filter the data.
        **kwds
            Additional keyword arguments to be passed to the super plot
            method.

        Returns
        -------
        Self
            The `BivariateUnivariateCharts` instance, for method 
            chaining.
        """
        self.hide_top_right_ax()
        for ax in self.axes:
            if ax == self.bivariate_ax:
                super().plot(
                    plotter,
                    kw_call=kw_call,
                    kw_where=kw_where,
                    on_last_axes=False,
                    **kwds)
            else:
                super().plot(SkipSubplot)
        return self
    
    def label(
            self,
            *,
            fig_title: str = '',
            sub_title: str = '',
            feature_label: str | bool | Tuple = '', 
            target_label: str | bool | Tuple = '', 
            info: bool | str = False,
            axes_titles: Tuple[str, ...] = (),
            rows: Tuple[str, ...] = (),
            cols: Tuple[str, ...] = (),
            row_title: str = '',
            col_title: str = '') -> Self:
        """Add labels and titles to the chart.

        This method sets various labels and titles for the chart,
        including figure title, subplot title, axis labels, row and
        column titles, and additional information.

        Parameters
        ----------
        fig_title : str, optional
            The main title for the entire figure, by default ''.
        sub_title : str, optional
            The subtitle for the entire figure, by default ''.
        feature_label : str | bool | None, optional
            The label for the feature variable (x-axis), by default ''.
            If set to True, the feature variable name will be used.
            If set to False or None, no label will be added.
        target_label : str | bool | None, optional
            The label for the target variable (y-axis), by default ''.
            If set to True, the target variable name will be used.
            If set to False or None, no label will be added.
        info : bool | str, optional
            Additional information to display on the chart. If True,
            the date and user information will be automatically added at
            the lower left corner of the figure. If a string is
            provided, it will be shown next to the date and user,
            separated by a comma. By default, no additional information
            is displayed.
        axes_titles : Tuple[str, ...]
            Title for each Axes, by default ()
        rows: Tuple[str, ...], optional
            The row labels of the figure, by default ().
        cols: Tuple[str, ...], optional
            The column labels of the figure, by default ().
        row_title : str, optional
            The title of the rows, by default ''.
        col_title : str, optional
            The title of the columns, by default ''.

        Returns
        -------
        Self
            The `BivariateUnivariateCharts` instance, for method
            chaining.

        Notes
        -----
        This method allows customization of chart labels and titles to
        enhance readability and provide context for the visualized data.
        """
        super().label(
            fig_title=fig_title,
            sub_title=sub_title,
            feature_label=(
                '', '',
                feature_label, ''),
            target_label=(
                '', '',
                target_label, ''),
            info=info,
            axes_titles=axes_titles,
            rows=rows,
            cols=cols,
            row_title=row_title,
            col_title=col_title)
        return self


class ProcessCapabilityAnalysisCharts(JointChart):
    """A class for creating process capability analysis charts.

    This class extends the `JointChart` class and provides methods for
    creating process capability analysis charts. It allows you to
    visualize process capability analysis data and perform various
    analysis tasks.


    Parameters
    ----------
    source : DataFrame
        The source data.
    target : str
        The target (dependant) variable.
    spec_limits : SpecLimits
        The specification limits for the process capability analysis.
    hue : str, optional
        The hue variable for the chart, by default ''.
    dist : scipy stats rv_continuous, optional
        The probability distribution use for creating feature data
        (the theoretical values). Default is 'norm'.
    error_values : tuple of float, optional
        If the process data may contain coded values for measurement 
        errors or similar, they can be specified here, 
        by default [].
    strategy : {'eval', 'fit', 'norm', 'data'}, optional
        Which strategy should be used to determine the control 
        limits (process spread):
        - `eval`: The strategy is determined according to the given 
        evaluate function. If none is given, the internal `evaluate`
        method is used.
        - `fit`: First, the distribution that best represents the 
        process data is searched for and then the agreed process 
        spread is calculated
        - norm: it is assumed that the data is subject to normal 
        distribution. The variation tolerance is then calculated as 
        agreement * standard deviation
        - data: The quantiles for the process variation tolerance 
        are read directly from the data.
        by default 'norm'
    agreement : float or int, optional
        Specify the tolerated process variation for which the 
        control limits are to be calculated. 
        - If int, the spread is determined using the normal 
        distribution agreement*σ, 
        e.g. agreement = 6 -> 6*σ ~ covers 99.75 % of the data. 
        The upper and lower permissible quantiles are then 
        calculated from this.
        - If float, the value must be between 0 and 1.This value is
        then interpreted as the acceptable proportion for the 
        spread, e.g. 0.9973 (which corresponds to ~ 6 σ)
        by default 6
    possible_dists : tuple of strings or rv_continous, optional
        Distributions to which the data may be subject. Only 
        continuous distributions of scipy.stats are allowed,
        by default DIST.COMMON
    stretch_figsize : bool | float | Tuple[float, float], optional
        If True, the height and width of the figure are stretched based 
        on the number rows and columns in the axes grid. If a float is 
        provided, the figure size is stretched by the given factor. If a 
        tuple of two floats is provided, the figure size is stretched by 
        the given factors for the x and y axis, respectively.
        by default False.
    
    Examples
    --------
    
    ```python
    import daspi as dsp

    df = dsp.load_dataset('drop_card')
    target = 'distance'

    chart = dsp.ProcessCapabilityAnalysisCharts(
            source=df,
            target=target,
            spec_limits=dsp.SpecLimits(0, float(df.loc[0, 'usl'])),
            hue='method'
        ).plot(
        ).stripes(
        ).label(
            fig_title='Process Capability Analysis',
            sub_title='Drop Card Experiment',
            target_label='Distance (cm)',
            info=True)
    ```

    ![ProcessCapabilityAnalysisChartsDocstringExample](https://raw.githubusercontent.com/j4ggr/DaSPi/main/docs/img/precast_process-capability-analysis-charts_example.png)
    """
    __slots__ = ('spec_limits', 'dist', 'kw_estim', 'n_groups')

    spec_limits: SpecLimits
    """The specification limits for the process capability analysis."""

    dist: rv_continuous | str
    """The probability distribution use for creating feature data
    (the theoretical values)."""

    kw_estim: Dict[str, Any]
    """Keyword arguments for the ProcessEstimator instances used for 
    calculating the capability indices."""

    n_groups: int
    """The number of hue groups in the source data."""

    def __init__(
            self,
            source: DataFrame,
            target: str,
            *,
            spec_limits: SpecLimits,
            hue: str = '',
            dist: rv_continuous | str = 'norm',
            error_values: Tuple[float, ...] = (),
            strategy: Literal['eval', 'fit', 'norm', 'data'] = 'norm',
            agreement: float | int = 6, 
            possible_dists: Tuple[str | rv_continuous, ...] = DIST.COMMON,
            stretch_figsize: bool | float | Tuple[float, float] = False,
            ) -> None:
        assert not spec_limits.both_unbounded, (
            'At least one specification limit must not be None')
        self.spec_limits = spec_limits
        self.dist = dist
        self.kw_estim = dict(
            error_values=error_values,
            strategy=strategy,
            agreement=agreement,
            possible_dists=possible_dists)
        self.n_groups = source.groupby(hue).ngroups if hue else 1
        filtered = source[~source[target].isin(error_values)]
        super().__init__(
            source=(filtered, filtered, filtered, source, source),
            target=target,
            feature='',
            hue=hue,
            mosaic=('rr', 'pd', 'ck',),
            height_ratios=[3, 3, 1],
            sharex='row',
            dodge=(False, False, False, True, True),
            target_on_y=(True, False, False, False, False),
            stretch_figsize=stretch_figsize)
    
    def processes(self) -> Dict[str, ProcessEstimator]:
        """Get a dictionary of ProcessEstimator instances that are used 
        for calculating the capability indices.
        
        Returns
        -------
        Dict[str, ProcessEstimator]
            A dictionary where:
            - Keys: The hue categories or the target name if no hue is 
              used.
            - Values: The corresponding ProcessEstimator instances.
        """
        if self.charts[0].hueing.n_used > 1:
            keys = self.charts[0].hueing.labels
        else:
            keys = iter((self.targets[0], ))

        process_instances = []
        for plot in self.plots:
            if isinstance(plot, CapabilityConfidenceInterval):
                if plot.kind == 'cpk':
                    try:
                        process_instance = next(iter(plot.processes.values()))
                        process_instances.append(process_instance)
                    except StopIteration:
                        warnings.warn(
                            'No process instances found in '
                            'CapabilityConfidenceInterval plot.')
        return dict(zip(keys, process_instances))
    
    def process_descriptions(self, exclude: Tuple[str, ...] = ()) -> DataFrame:
        """Get a DataFrame containing the descriptions of each process.
        
        This method generates a DataFrame that contains the descriptions
        of each process, excluding any specified columns.
        
        Parameters
        ----------
        exclude : Tuple[str, ...], optional
            Attributes to exclude from the summary statistics,
            by default ().
        
        Returns
        -------
        DataFrame
            A DataFrame containing the descriptions of each process."""

        descriptions = [
            p.describe(exclude=exclude, colname=n)
            for n, p in self.processes().items()]
        df_p = pd.concat(descriptions, axis=1)
        return df_p

    def plot(self, **kwds_cpi) -> Self: # type: ignore
        """Plot the process capability analysis charts.
        
        This method plots the process capability analysis charts, including
        scatter plots, probability density functions, Gaussian kernel
        density estimation, and capability confidence intervals.

        Parameters
        ----------
        kwds_cpi : dict
            Additional keyword arguments for the capability confidence
            interval plot.

        Returns
        -------
        Self
            The `ProcessCapabilityAnalysisCharts` instance, for method
            chaining.
        """
        _kwds_cpi: Dict[str, Any] = dict(
            n_groups=self.n_groups,
            spec_limits=self.spec_limits,
            show_center=True,
            bars_same_color=True,
            hide_axis='feature',
            visible_spines='target',
            kw_estim=self.kw_estim) | kwds_cpi
        super().plot(Scatter)
        super().plot(Probability, dist=self.dist)
        super().plot(GaussianKDE, hide_axis='feature', visible_spines='target')
        super().plot(CapabilityConfidenceInterval, kind='cpk', **_kwds_cpi)
        if self.spec_limits.is_unbounded:
            super().plot(HideSubplot)
        else:
            super().plot(CapabilityConfidenceInterval, kind='cp', **_kwds_cpi)

        return self
    
    def stripes(
            self,
            *,
            mean: bool = True,
            median: bool = True,
            control_limits: bool = True
            ) -> Self: # type: ignore
        """This method adds stripes to the process capability analysis 
        charts, including mean, median, control limits and specification 
        limits.

        Parameters
        ----------
        mean : bool, optional
            Whether to add a line for the mean, by default True.
        median : bool, optional
            Whether to add a line for the median
            (if applicable), by default True.
        control_limits : bool, optional
            Whether to add lines for the control limits
            (if applicable), by default True.
        
        Returns
        -------
        Self
            The `ProcessCapabilityAnalysisCharts` instance, for method
            chaining.
        """
        def flags(x: bool) -> Generator[bool, Any, None]:
            for flag in (x, False, x, False, False):
                yield flag
        
        super().stripes(
            mean=tuple(flags(mean)),
            median=tuple(flags(median)),
            control_limits=tuple(flags(control_limits)),
            spec_limits=tuple(
                self.spec_limits if f else SpecLimits() for f in flags(True)))
        return self

    def label( # type: ignore
            self,
            *,
            fig_title: str = '',
            sub_title: str = '',
            feature_label: Tuple[str, ...] = (),
            target_label: str = '',
            info: bool | str = False,
            ) -> Self:
        """Label the process capability analysis charts.
        
        This method labels the process capability analysis charts with
        custom labels and titles.

        Parameters
        ----------
        fig_title : str, optional
            The title of the figure, by default ''.
        sub_title : str, optional
            The subtitle of the figure, by default ''.
        feature_label : Tuple[str, ...], optional
            The labels for the features. If not provided, default labels
            will be used. The tuple should contain five elements. By
            default, the labels are:
            ('Observation order',
            '', 'Quantiles of standard normal distribution', 
            '', '')
        target_label : str, optional
            The label for the target variable, by default ''.
        info : bool | str, optional
            Whether to display information about the chart. If True, the
            information will be displayed. If a string is provided, it
            will be used as the title of the information box. By default,
            the information is not displayed.

        Returns
        -------
        Self
            The `ProcessCapabilityAnalysisCharts` instance, for method
            chaining.
        """
        if not target_label:
            target_label = self.targets[0]
        _target_label = 3 * (target_label, ) + (STR['cpk'], STR['cp'])
        if not feature_label:
            feature_label = (
                STR['charts_flabel_observed'], 
                STR['charts_flabel_quantiles'], '', '', '')
        super().label(
            fig_title=fig_title,
            sub_title=sub_title,
            target_label=_target_label,
            feature_label=feature_label,
            info=info)
        return self


class GageStudyCharts(JointChart):
    """
    A class for visualizing the parameters of a GageStudy model.

    Parameters
    ----------
    model : GageStudyModel
        The GageStudy model whose parameters will be visualized.
    dist : scipy stats rv_continuous, optional
        The probability distribution use for creating feature data
        (the theoretical values). Default is 'norm'.
    stretch_figsize : bool | float | Tuple[float, float], optional
        If True, the height and width of the figure are stretched based 
        on the number rows and columns in the axes grid. If a float is 
        provided, the figure size is stretched by the given factor. If a 
        tuple of two floats is provided, the figure size is stretched by 
        the given factors for the x and y axis, respectively.
        by default False.
    
    Examples
    --------

    Run the following command in a jupyter notebook to get the html
    output of `gage` or you can also use `print(repr(gage))` instead:

    ```python
    import daspi as dsp
    df = dsp.load_dataset('grnr_layer_thickness')
    gage = dsp.GageStudyModel(
        source=df,
        target='result_gage',
        reference='reference',
        u_cal=df['U_cal'][0],
        tolerance=df['tolerance'][0],
        resolution=df['resolution'][0],
        bias_corrected=True,)
    chart = dsp.GageStudyCharts(gage, stretch_figsize=1.5).plot().stripes().label()
    gage
    ```

    ![GageStudyChartsDocstringExample](https://raw.githubusercontent.com/j4ggr/DaSPi/main/docs/img/precast_gage-study-charts_example.png)
    """
    
    __slots__ = (
        'model',
        'dist',
        'has_multiple_references')

    model: GageStudyModel
    """The GageStudy model whose parameters are visualized."""
    dist: rv_continuous | str
    """The probability distribution use for creating feature data
    (the theoretical values)."""
    has_multiple_references: bool
    """Whether the GageStudy model has multiple references."""

    def __init__(
            self,
            model: GageStudyModel,
            *,
            dist: rv_continuous | str = 'norm',
            stretch_figsize: bool | float | Tuple[float, float] = False
            ) -> None:
        self.model = model
        source = model.source.copy()
        reference = model.reference
        df_cap = model.capabilities()
        cap_target = ANOVA.CAPABILITY_COLNAMES[0]
        cap_features = (df_cap.reset_index(drop=False).columns[0], )*5
        self.dist = dist
        self.has_multiple_references = source[reference].nunique() > 1
        
        if self.has_multiple_references:
            feature = (reference, '', '', *cap_features)
            hue = tuple('' if f == reference else reference for f in feature)
            categorical_feature = (
                True, False, False, True, True, True, True, True)
        else:
            feature = (STR['charts_flabel_observed'], '', '', *cap_features)
            hue = ('', ) * 8
            categorical_feature = (
                False, False, False, True, True, True, True, True)
            source[feature[0]] = [i + 1 for i in range(len(source))]

        dfs = tuple(
            df_cap.loc[[r], :].reset_index(drop=False)
            for r in ANOVA.CAPABILITY_ROWS)

        super().__init__(
            source=(source, ) * 3 + dfs,
            target=(model.target, ) * 3 + (cap_target, ) * 5,
            feature=feature,
            hue=hue,
            mosaic=('ooooo', 'ppddd', 'gkrtb'),
            target_on_y=(True, False, False, True, True, True, True, True),
            categorical_feature=categorical_feature,
            stretch_figsize=stretch_figsize)
    
    def plot(self) -> Self: # type: ignore
        if self.has_multiple_references:
            super().plot(
                CategoricalObservation, show_line=True, show_scatter=True)
        else:
            super().plot(Line)
            super().plot(Scatter, on_last_axes=True)
        super().plot(Probability, dist=self.dist)
        super().plot(GaussianKDE, hide_axis='feature', visible_spines='target')
        super().plot(Bar)
        super().plot(Bar)
        super().plot(Bar)
        super().plot(Bar)
        super().plot(Bar)
        return self
    
    def stripes(self) -> Self:
        df_cap = self.model.capabilities()

        for idx in range(3):
            self.charts[idx].stripes(
                mean=True)

        idx = -5
        for name, limit in df_cap[ANOVA.CAPABILITY_COLNAMES[1]].items():
            if name in ANOVA.CAPABILITY_ROWS[2:4]:
                accepted = StripeSpan(
                    label=STR['accepted'],
                    lower_position=0,
                    upper_position=limit,
                    orientation='horizontal',
                    color=COLOR.GOOD,)
                rejected = StripeSpan(
                    label=STR['rejected'],
                    orientation='horizontal',
                    lower_position=limit,
                    upper_position=1,
                    color=COLOR.BAD,)
            else:
                rejected = StripeSpan(
                    label=STR['rejected'],
                    lower_position=0,
                    upper_position=limit,
                    orientation='horizontal',
                    color=COLOR.BAD,)
                accepted = StripeSpan(
                    label=STR['accepted'],
                    lower_position=limit,
                    upper_position=5,
                    orientation='horizontal',
                    color=COLOR.GOOD,)
            self.charts[idx].stripes([accepted, rejected])
            idx += 1
        return self

    def label(self, info: bool | str = False, **kwds) -> Self: # type: ignore
        """Adds titles and labels to the charts generated by the 
        `plot()` method.
        
        Parameters
        ----------
        info : bool | str, optional
            If `True`, the method will add an informative subtitle to 
            the chart. If a string is provided, it will be used as the 
            subtitle, by default False.
        **kwds
            Additional keyword arguments to be passed to the `label()`
            method of the `JointChart` instance.
        
        Returns
        -------
        Self
            The `ParameterRelevanceCharts` instance, for method chaining.
        """      
        labels: Dict[str, Any] = dict(
                fig_title=STR['gstudycharts_fig_title'],
                sub_title=STR['gstudycharts_sub_title'],
                target_label=(
                    self.model.target,
                    self.model.target,
                    self.model.target,
                    '', '', '', ''),
                feature_label=(
                    STR['charts_flabel_observed'],
                    STR['charts_flabel_quantiles'],
                    '', '', '', '', ''),
                info = info
            ) | kwds
        super().label(**labels)

        for idx, lim in zip((-5, -4, -3, -2, -1), (5, 5, 0.2, 0.6, 0.2)):
            ax = self.axes[idx]
            if idx in (-3, -2,):
                ax.yaxis.set_major_formatter(PercentFormatter(xmax=1))
            ax.set(ylim=(0, lim))
        return self


class GageRnRCharts(JointChart):
    """
    A class for visualizing the parameters of a GageRnR model.

    Parameters
    ----------
    rnr_model : GageRnRModel
        The GageRnR model whose parameters will be visualized.
    spread_accepted_limit: float, optional
        The maximum value for the the spread proportion where it is 
        accepted. Default is 0.1 (10 %).
    spread_rejected_limit: float, optional
        The minimum limit where the spread proportion is rejected.
        Default is 0.3 (30 %).
    u_accepted_limit: float, optional
        The maximum value for the the u proportion where it is accepted.
        Default is 0.15 (15 %).
    stretch_figsize : bool | float | Tuple[float, float], optional
        If True, the height and width of the figure are stretched based 
        on the number rows and columns in the axes grid. If a float is 
        provided, the figure size is stretched by the given factor. If a 
        tuple of two floats is provided, the figure size is stretched by 
        the given factors for the x and y axis, respectively.
        by default False.
    
    Examples
    --------
    ```python
    import daspi as dsp

    df = dsp.load_dataset('grnr_layer_thickness')
    gage = dsp.GageStudyModel(
        source=df,
        target='result_gage',
        reference='reference',
        u_cal=df['U_cal'][0],
        tolerance=df['tolerance'][0],
        resolution=df['resolution'][0])
    model = dsp.GageRnRModel(
        source=df,
        target='result_rnr',
        part='part',
        gage=gage,
        u_av='operator')

    chart = dsp.GageRnRCharts(model, stretch_figsize=True
        ).plot(
        ).stripes(
        ).label()
    model
    ```

    ![GageRnRChartsDocstringExample](https://raw.githubusercontent.com/j4ggr/DaSPi/main/docs/img/precast_gage-rnr-charts_example.png)
    """
    
    __slots__ = (
        'model', 
        'spread_accepted_limit', 
        'spread_rejected_limit',
        'u_accepted_limit')

    model: GageRnRModel
    """The GageRnR model whose parameters are visualized."""

    spread_accepted_limit: float
    """The maximum value for the the spread proportion where it is 
    accepted (10 %)."""

    spread_rejected_limit: float
    """The minimum limit where the spread proportion is rejected (30 %)."""

    u_accepted_limit: float
    """The minimum limit where the measuremen unsertainty proportion
     is rejected (15 %)."""

    def __init__(
            self,
            rnr_model: GageRnRModel,
            *,
            spread_accepted_limit: float = 0.1,
            spread_rejected_limit: float = 0.3,
            u_accepted_limit: float = 0.15,
            stretch_figsize: bool | float | Tuple[float, float] = False,
            ) -> None:
        self.model = rnr_model
        self.spread_accepted_limit = spread_accepted_limit
        self.spread_rejected_limit = spread_rejected_limit
        self.u_accepted_limit = u_accepted_limit

        target = self.model.target
        feature_spread = STR['lm_table_rnr_source']
        hue_spread = STR['rnrcharts_spread_proportions']
        columns_map = {
            v: k for k, v in 
            (self.model.target_map | self.model.feature_map).items()}
        
        df_model = (self.model.data
            .copy()
            .rename(columns=columns_map)
            [list(columns_map.values())])
        
        df_mean = (df_model
                .groupby([self.av_gv, self.part], observed=True)
                [target]
                .mean()
                .to_frame()
                .reset_index(drop=False))
        
        df_span = (df_model
                .groupby([self.av_gv, self.part], observed=True)
                [target]
                .agg(['min', 'max'])
                .diff(axis=1)
                .drop(columns=['min'])
                .rename(columns={'max': target})
                .reset_index(drop=False))
        
        av_gv = ANOVA.AV if self.model.has_operator else ANOVA.GV
        df_rnr = (self.model
            .rnr()
            .loc[[ANOVA.RNR, ANOVA.EV, av_gv], :]
            .reset_index(drop=False)
            .rename(columns={'index':feature_spread}))
        datas = []
        for col in ANOVA.RNR_COLNAMES[-2:]:
            data = df_rnr[[feature_spread, col]].rename(columns={col: target})
            data[hue_spread] = col
            datas.append(data)
        df_rnr = pd.concat(datas, axis=0, ignore_index=True)

        df_u = (self.model
            .df_u.loc[['RE', 'MS', 'MP'], ['Q']]
            .reset_index(drop=False)
            .rename(columns={'index':feature_spread, 'Q': target}))
            
        super().__init__(
            source=(df_model, df_mean, df_span, df_span, df_rnr, df_u),
            target=target,
            feature=(
                self.part, self.av_gv, self.part, self.av_gv, 
                feature_spread, feature_spread),
            hue=(self.av_gv, '', self.av_gv, '', hue_spread, ''),
            ncols=2,
            nrows=3,
            stretch_figsize=stretch_figsize,
            dodge=(True, False, True, False, True, False),
            target_on_y=(True, True, True, True, False, False),
            categorical_feature=True)
        
        self.charts[-2].hueing._categories = (
            COLOR.SPECIAL_AREA1, COLOR.SPECIAL_AREA2)
        self.axes[0].sharey(self.axes[1])
        self.axes[2].sharey(self.axes[3])

    @property
    def part(self) -> str:
        """Column name of the part (unit under test) variable 
        (read-only)."""
        return self.model.part
    
    @property
    def av_gv(self) -> str:
        """Column name of the variable that identifies the operator for 
        type 2 Gage R&R or the gage variate for type 3 Gage R&R
        (read-only)."""
        value = ANOVA.AV if self.model.has_operator else ANOVA.GV
        av_gv = {v: k for k, v in self.model.u_map.items()}.get(value, None)
        assert av_gv is not None, (
            'The provided gage study model does not have an operator or '
            'gage variate.')
        return av_gv

    def plot(self) -> Self:
        """Plot the GageRnR charts."""

        # top Left
        super().plot(
            Scatter)
        super().plot(
            CenterLocation,
            on_last_axes=True,
            show_center=True,
            show_line=True,
            kw_call=dict(marker='_'))

        # top right
        super().plot(
            ParallelCoordinate,
            identity=self.part,
            kw_call=dict(color=DEFAULT.PLOTTING_COLOR, alpha=COLOR.MARKER_ALPHA))

        # mid left
        super().plot(
            Stem,
            bottom=0,
            base_color=COLOR.TRANSPARENT)
        
        # mid right
        super().plot(
            Beeswarm)
        super().plot(
            CenterLocation,
            show_line=True,
            show_center=False,
            on_last_axes=True,
            kw_call=dict(marker='_'))
        super().plot(
            MeanTest,
            show_center=True,
            on_last_axes=True,
            kw_call=dict(kw_center=dict(marker='_', s=10)))
        
        # bottom left
        super().plot(
            Bar)
        
        # bottom right
        super().plot(
            Bar)

        return self
    
    def stripes(self) -> Self: # type: ignore
        """Adds lines for the limits for the spread and uncertainties 
        proportions."""
        rnr_accepted = StripeSpan(
            label=STR['accepted'],
            lower_position=0,
            upper_position=self.spread_accepted_limit,
            orientation='vertical',
            color=COLOR.GOOD,)
        rnr_borderline = StripeSpan(
            label=STR['borderline'],
            lower_position=self.spread_accepted_limit,
            upper_position=self.spread_rejected_limit,
            orientation='vertical',
            color=COLOR.ANOMALY,)
        rnr_rejected = StripeSpan(
            label=STR['rejected'],
            lower_position=self.spread_rejected_limit,
            upper_position=1,
            orientation='vertical',
            color=COLOR.BAD,)
        self.charts[-2].stripes([rnr_accepted, rnr_borderline, rnr_rejected])
        
        u_accepted = StripeSpan(
            label=STR['accepted'],
            lower_position=0,
            upper_position=self.u_accepted_limit,
            orientation='vertical',
            color=COLOR.GOOD,)
        u_rejected = StripeSpan(
            label=STR['rejected'],
            lower_position=self.u_accepted_limit,
            upper_position=1,
            orientation='vertical',
            color=COLOR.BAD,)
        self.charts[-1].stripes([u_accepted, u_rejected])
        return self

    def label(self, info: bool | str = False, **kwds) -> Self: # type: ignore
        """Adds titles and labels to the charts generated by the 
        `plot()` method.
        
        Parameters
        ----------
        info : bool | str, optional
            If `True`, the method will add an informative subtitle to 
            the chart. If a string is provided, it will be used as the 
            subtitle, by default False.
        **kwds
            Additional keyword arguments to be passed to the `label()`
            method of the `JointChart` instance.
        
        Returns
        -------
        Self
            The `ParameterRelevanceCharts` instance, for method chaining.
        """      
        labels: Dict[str, Any] = dict(
                fig_title=STR['rnrcharts_fig_title'],
                sub_title=STR['rnrcharts_sub_title'],
                target_label=(
                    self.model.target, '', STR['data_range'], '', 
                    STR['rnrcharts_spread_proportions'], STR['rnrcharts_suitability']),
                feature_label=(
                    True, True, True, True, True, False),
                info = info
            ) | kwds
        super().label(**labels)

        for ax in (self.axes[-2], self.axes[-1]):
            ax.xaxis.set_major_formatter(PercentFormatter(xmax=1))
            ax.set(xlim=(0, 0.5))
        return self


class PairwiseMatrixCharts(JointChart):
    """A class for creating pairwise matrix charts.

    This class extends the `JointChart` class and provides methods for
    creating pairwise matrix charts. It allows you to visualize pairwise
    relationships between features in a dataset.

    Parameters
    ----------
    source : DataFrame
        The source data.
    features : Tuple[str, ...]
        The features for which the pairwise matrix charts will be 
        created.
    hue : str, optional
        The hue variable for the charts, by default ''.
    stretch_figsize : bool | float | Tuple[float, float], optional
        If True, the height and width of the figure are stretched based
        on the number rows and columns in the axes grid. If a float is
        provided, the figure size is stretched by the given factor. If a
        tuple of two floats is provided, the figure size is stretched by
        the given factors for the x and y axis, respectively.
        by default False.
    
    Examples
    --------
    ```python
    import daspi as dsp
    df = dsp.load_dataset('penguins')
    features = (
        'Culmen Length (mm)',
        'Culmen Depth (mm)',
        'Flipper Length (mm)',
        'Body Mass (g)')
    chart = dsp.PairwiseMatrixCharts(
        source=df,
        features=features,
        hue='Species',
        stretch_figsize=True
        ).plot(
        ).label(
            fig_title='Pairwise Matrix Charts',
            sub_title='Penguins Dataset',
            target_label=True,
            feature_label=True,
            info=True
        )
    ```

    ![PairwiseMatrixChartsDocstringExample](https://raw.githubusercontent.com/j4ggr/DaSPi/main/docs/img/precast_pairwise-matrix-charts_example.png)
    """
    __slots__ = ('_diag_axes',)

    _diag_axes: Tuple[Axes, ...]
    """A tuple containing the additional diagonal axes of the pairwise matrix 
    charts."""
    
    def __init__(
            self,
            source: DataFrame,
            features: Tuple[str, ...],
            hue: str = '',
            stretch_figsize: bool | float | Tuple[float, float] = False,
            ) -> None:
        
        n = len(features)
        targets = tuple(f for f in features for _ in range(n))
        features = tuple(f for _ in range(n) for f in features)
        equals = tuple(f == t for f, t in zip(features, targets))
        super().__init__(
            source=source,
            target=targets,
            feature=features,
            hue=hue,
            ncols=n,
            nrows=n,
            sharex='col',
            sharey='row',
            stretch_figsize=stretch_figsize)

        self._diag_axes = tuple(
            ax.twiny() for ax, eq in zip(self.axes, equals) if eq)
    
    @check_label_order
    def plot(
            self,
            *,
            diag_plot: Type[Plotter] | None = GaussianKDE,
            lower_plot: Type[Plotter] | None = LinearRegressionLine,
            upper_plot: Type[Plotter] | None = GaussianKDEContour,
            kw_diag: Dict[str, Any] | None = {
                'visible_spines': 'target',
                'hide_axis': 'both',},
            kw_lower: Dict[str, Any] | None = None,
            kw_upper: Dict[str, Any] | None = None,
            ) -> Self:
        """Plot the pairwise matrix charts.

        Parameters
        ----------
        diag_plot : Type[Plotter] | None, optional
            The type of plot to use for the diagonal subplots. If None, 
            the diagonal subplots will be skipped. Default is 
            :class:`GaussianKDE`.
        lower_plot : Type[Plotter] | None, optional
            The type of plot to use for the lower triangle subplots. If
            None, the lower triangle subplots will be skipped. Default 
            is :class:`LinearRegressionLine`.
        upper_plot : Type[Plotter] | None, optional
            The type of plot to use for the upper triangle subplots. If
            None, the upper triangle subplots will be skipped. Default 
            is :class:`GaussianKDEContour`.
        kw_diag : Dict[str, Any] | None, optional
            Additional keyword arguments to be passed to the plotter for
            the diagonal subplots. Default is 
            :obj:`{'visible_spines': 'target', 'hide_axis': 'feature'}`.
        kw_lower : Dict[str, Any] | None, optional
            Additional keyword arguments to be passed to the plotter for
            the lower triangle subplots. Default is None.
        kw_upper : Dict[str, Any] | None, optional
            Additional keyword arguments to be passed to the plotter for
            the upper triangle subplots. Default is None.
        """
        kw_diag = kw_diag or dict()
        kw_lower = kw_lower or dict()
        kw_upper = kw_upper or dict()
        for row, diag_ax in zip(range(self.axes.shape[0]), self._diag_axes):
            for col in range(self.axes.shape[1]):
                kwds = dict()
                plotter = SkipSubplot
                chart = self._next_chart_()
                if diag_plot is not None and row == col:
                    plotter = diag_plot
                    self.axes._ax = diag_ax
                    kwds = kw_diag
                elif lower_plot is not None and row > col:
                    plotter = lower_plot
                    kwds = kw_lower
                elif upper_plot is not None and row < col:
                    plotter = upper_plot
                    kwds = kw_upper
                chart.plot(plotter, **kwds)
        return self
    
    def label(# type: ignore
            self,
            *,
            fig_title: str = '',
            sub_title: str = '',
            feature_label: bool | Tuple[str, ...] = True,
            target_label: bool | Tuple[str, ...] = True,
            info: bool | str = False,
            ) -> Self:
        """Add labels and titles to the chart.

        This method sets various labels and titles for the chart, 
        including figure title, subplot title and additional information.
        
        Parameters
        ----------
        fig_title : str, optional
            The main title for the entire figure, by default ''.
        sub_title : str, optional
            The subtitle for the entire figure, by default ''.
        feature_label : bool | Tuple[str, ...], optional
            The label for the feature variable (x-axis). If set to True, 
            the feature variable names will be used. If set to False or 
            None, no label will be added. If a tuple of strings is 
            provided, it will be used as the labels. Default is True.
        target_label : bool | Tuple[str, ...], optional
            The label for the target variable (y-axis), by default ''. 
            If set to True, the target variable name will be used. 
            If set to False or None, no label will be added.
            If a tuple of strings is provided, it will be used as the 
            labels. Default is True.
        info : bool | str, optional
            Additional information to display on the chart. If True, 
            the date and user information will be automatically added 
            at the lower left corner of the figure. If a string is 
            provided, it will be shown next to the date and user, 
            separated by a comma. By default, no additional information 
            is displayed.
        """
        if feature_label is True:
            feature_label = self.features

        if target_label is True:
            target_label = self.targets

        return super().label(
            fig_title=fig_title,
            sub_title=sub_title,
            feature_label=feature_label,
            target_label=target_label,
            info=info,)



import sys
import pytest

import numpy as np
import pandas as pd

from typing import Any
from typing import Dict
from pytest import approx
from pathlib import Path
from pandas.core.frame import DataFrame

sys.path.append(Path(__file__).parent.resolve()) # type: ignore

from daspi.statistics.hypothesis import *

source = Path(__file__).parent/'data'
KW_READ: Dict[str, Any] = dict(sep=';', index_col=0)

df_dist10: DataFrame = pd.read_csv(
    source/f'dists_10-samples.csv', skiprows=1, nrows=10, **KW_READ)
df_valid10: DataFrame = pd.read_csv(
    source/f'dists_10-samples.csv', skiprows=14, **KW_READ)
df_dist25: DataFrame = pd.read_csv(
    source/f'dists_25-samples.csv', skiprows=1, nrows=25, **KW_READ)
df_valid25: DataFrame = pd.read_csv(
    source/f'dists_25-samples.csv', skiprows=29, **KW_READ)


class TestChunker:

    samples = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])

    def test_with_valid_input(self) -> None:
        sections = 3
        result = list(chunker(self.samples, sections))
        assert len(result) == sections
        assert np.array_equal(result[0], np.array([1, 2, 3, 4]))
        assert np.array_equal(result[1], np.array([5, 6, 7]))
        assert np.array_equal(result[2], np.array([8, 9, 10]))

    def test_with_single_section(self) -> None:
        sections = 1
        result = list(chunker(self.samples, sections))
        assert len(result) == sections
        assert np.array_equal(result[0], self.samples)

    def test_with_single_sample(self) -> None:
        sections = 2
        samples = [1]
        result = list(chunker(samples, sections))
        assert len(result) == sections
        assert np.array_equal(result[0], samples)
        assert result[1].size == 0

    def test_with_zero_sections(self) -> None:
        sections = 0
        with pytest.raises(AssertionError):
            list(chunker(self.samples, sections))

    def test_with_negative_sections(self) -> None:
        sections = -2
        with pytest.raises(AssertionError):
            list(chunker(self.samples, sections))

    def test_with_non_integer_sections(self) -> None:
        sections = 2.5
        with pytest.raises(AssertionError):
            list(chunker(self.samples, sections)) # type: ignore

# --- GROUPED TESTS FOR hypothesis.py COVERAGE ---
class TestHypothesisFunctions:
    def test_anderson_darling_test(self):
        from daspi.statistics.hypothesis import anderson_darling_test
        data = np.random.normal(0, 1, 100)
        p, stat = anderson_darling_test(data)
        assert isinstance(p, float)
        assert isinstance(stat, float)
        assert 0 <= p <= 1

    def test_all_normal(self):
        from daspi.statistics.hypothesis import all_normal
        a = np.random.normal(0, 1, 50)
        b = np.random.normal(0, 1, 50)
        assert all_normal(a, b)
        assert all_normal(a, b, p_threshold=0.01) in [True, False]
        with pytest.raises(AssertionError):
            all_normal(a, b, p_threshold=1.5)

    def test_kolmogorov_smirnov_test(self):
        from daspi.statistics.hypothesis import kolmogorov_smirnov_test
        data = np.random.normal(0, 1, 100)
        p, D, params = kolmogorov_smirnov_test(data, 'norm')
        assert isinstance(p, float)
        assert isinstance(D, float)
        assert isinstance(params, tuple)
        assert 0 <= p <= 1

    def test_f_test(self) -> None:
        from daspi.statistics.hypothesis import f_test
        a = np.random.normal(0, 1, 30)
        b = np.random.normal(0, 1, 30)
        p, F = f_test(a, b)
        assert isinstance(p, float)
        assert isinstance(F, float)
        assert 0 <= p <= 1

    def test_t_test(self) -> None:
        from daspi.statistics.hypothesis import t_test
        a = np.random.normal(0, 1, 30)
        p, t, df = t_test(a)
        assert isinstance(p, float)
        assert isinstance(t, float)
        assert isinstance(df, int)
        assert 0 <= p <= 1

    def test_levene_test(self) -> None:
        from daspi.statistics.hypothesis import levene_test
        a = np.random.normal(0, 1, 30)
        b = np.random.normal(0, 1, 30)
        p, L = levene_test(a, b)
        assert isinstance(p, float)
        assert isinstance(L, float)
        assert 0 <= p <= 1

    def test_variance_stability_test(self) -> None:
        from daspi.statistics.hypothesis import variance_stability_test
        a = np.random.normal(0, 1, 30)
        p, L = variance_stability_test(a, n_sections=3)
        assert isinstance(p, float)
        assert isinstance(L, float)
        assert 0 <= p <= 1

    def test_mean_stability_test(self) -> None:
        from daspi.statistics.hypothesis import mean_stability_test
        a = np.random.normal(0, 1, 30)
        p, stat = mean_stability_test(a, n_sections=3)
        assert isinstance(p, float)
        assert isinstance(stat, float)
        assert 0 <= p <= 1

    def test_position_test(self) -> None:
        from daspi.statistics.hypothesis import position_test
        a = np.random.normal(0, 1, 30)
        b = np.random.normal(0, 1, 30)
        p, stat, test = position_test(a, b)
        assert isinstance(p, float)
        assert isinstance(stat, float)
        assert isinstance(test, str)
        assert 0 <= p <= 1

    def test_variance_test(self) -> None:
        from daspi.statistics.hypothesis import variance_test
        a = np.random.normal(0, 1, 30)
        b = np.random.normal(0, 1, 30)
        p, stat, test = variance_test(a, b)
        assert isinstance(p, float)
        assert isinstance(stat, float)
        assert isinstance(test, str)
        assert 0 <= p <= 1

    def test_proportions_test(self) -> None:
        from daspi.statistics.hypothesis import proportions_test
        p, stat, test = proportions_test(5, 10, 7, 12)
        assert isinstance(p, float)
        assert isinstance(stat, float)
        assert isinstance(test, str)
        assert 0 <= p <= 1

    def test_kurtosis_test(self) -> None:
        from daspi.statistics.hypothesis import kurtosis_test
        a = np.random.normal(0, 1, 30)
        p, stat = kurtosis_test(a)
        assert isinstance(p, float)
        assert isinstance(stat, float)
        assert 0 <= p <= 1

    def test_skew_test(self) -> None:
        from daspi.statistics.hypothesis import skew_test
        a = np.random.normal(0, 1, 30)
        p, stat = skew_test(a)
        assert isinstance(p, float)
        assert isinstance(stat, float)
        assert 0 <= p <= 1

    def test_ensure_generic(self) -> None:
        from daspi.statistics.hypothesis import ensure_generic
        dist = ensure_generic('norm')
        assert hasattr(dist, 'cdf')
        import scipy.stats as stats
        dist2 = ensure_generic(stats.norm)
        assert hasattr(dist2, 'cdf')
        with pytest.raises(AttributeError):
            ensure_generic('not_a_dist')


import sys
import pytest

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from numpy.typing import NDArray

from typing import Any
from typing import Generator
from pathlib import Path

sys.path.append(Path(__file__).parent.resolve()) # type: ignore

from daspi import JointChart
from daspi import SingleChart
from daspi import MultivariateChart
from daspi import SpecLimits
from daspi.plotlib.plotter import Scatter
from daspi.plotlib.appearance import positions_of_shared_axes

class TestSharingAxesFunctions:
    """tests for get_shared_axes and positions_of_shared_axes functions"""
    @pytest.fixture
    def shared_axes(self) -> Generator[NDArray, Any, None]:
        fig, axs = plt.subplots(2, 2, sharex='col', sharey='row')
        yield axs
        plt.close(fig)
    
    def test_position_types(self, shared_axes: NDArray) -> None:
        ax = shared_axes[0, 0]
        positions = positions_of_shared_axes(ax, 'x')
        assert isinstance(positions, list)
        assert all(isinstance(p, int) for p in positions)

        positions = positions_of_shared_axes(ax, 'y')
        assert isinstance(positions, list)
        assert all(isinstance(p, int) for p in positions)
    
    def test_axes_order(self, shared_axes: NDArray) -> None:
        figure_axes = shared_axes[0, 0].figure.axes
        assert figure_axes[0] == shared_axes[0, 0]
        assert figure_axes[1] == shared_axes[0, 1]
        assert figure_axes[2] == shared_axes[1, 0]
        assert figure_axes[3] == shared_axes[1, 1]

    def test_positions_shared_x_axis(self, shared_axes: NDArray) -> None:
        positions = positions_of_shared_axes(shared_axes[0, 0], 'x')
        assert positions == [0, 2]
        positions = positions_of_shared_axes(shared_axes[0, 1], 'x')
        assert positions == [1, 3]
        positions = positions_of_shared_axes(shared_axes[1, 0], 'x')
        assert positions == [0, 2]
        positions = positions_of_shared_axes(shared_axes[1, 1], 'x')
        assert positions == [1, 3]

    def test_positions_shared_y_axis(self, shared_axes: NDArray) -> None:
        positions = positions_of_shared_axes(shared_axes[0, 0], 'y')
        assert positions == [0, 1]
        positions = positions_of_shared_axes(shared_axes[0, 1], 'y')
        assert positions == [0, 1]
        positions = positions_of_shared_axes(shared_axes[1, 0], 'y')
        assert positions == [2, 3]
        positions = positions_of_shared_axes(shared_axes[1, 1], 'y')
        assert positions == [2, 3]

    def test_positions_no_shared_axes(self) -> None:
        fig, ax = plt.subplots(1, 1)
        positions = positions_of_shared_axes(ax, 'x')
        assert isinstance(positions, list)
        assert len(positions) == 1
        assert positions == [0]
        plt.close(fig)

    def test_positions_empty_figure(self) -> None:
        fig = plt.figure()
        ax = fig.add_subplot(111)
        positions = positions_of_shared_axes(ax, 'x')
        assert isinstance(positions, list)
        assert len(positions) == 1
        plt.close(fig)


class TestSingleChart:

    @pytest.fixture
    def sample_data(self) -> pd.DataFrame:
        return pd.DataFrame({
            'target': [1, 2, 3, 4, 5],
            'feature': ['A', 'B', 'C', 'D', 'E']
        })

    def test_initialization(self, sample_data: pd.DataFrame) -> None:
        chart = SingleChart(sample_data, 'target', 'feature')
        assert chart.source.equals(sample_data)
        assert chart.target == 'target'
        assert chart.feature == 'feature'
        assert chart.target_on_y == True

    def test_initialization_target_on_x(self, sample_data: pd.DataFrame) -> None:
        chart = SingleChart(sample_data, 'target', 'feature', target_on_y=False)
        assert chart.target_on_y == False

    def test_initialization_with_kwds(self, sample_data: pd.DataFrame) -> None:
        chart = SingleChart(sample_data, 'target', 'feature', sharex='col', sharey='row')
        assert chart.axes.sharex == 'col'
        assert chart.axes.sharey == 'row'


class TestJointChart:

    @pytest.fixture
    def sample_data(self) -> pd.DataFrame:
        return pd.DataFrame({
            'x': [1, 2, 3, 4, 5],
            'y': [2, 4, 6, 8, 10],
            'z': [1.1, 0.9, 1.0, 1.2, 0.8],
            'category': ['A', 'B', 'A', 'B', 'A'],
        })

    def test_initialization(self, sample_data: pd.DataFrame) -> None:
        chart = JointChart(sample_data, feature='x', target='y', ncols=1, nrows=1)
        assert isinstance(chart, JointChart)
        assert chart.source.equals(sample_data)
        assert chart.features == ('x',)
        assert chart.targets == ('y',)
        assert chart.target_on_ys == (True,)

    def test_hue(self, sample_data: pd.DataFrame) -> None:
        chart = JointChart(
            sample_data, feature='x', target='y', hue='category',
            ncols=2, nrows=3)
        assert chart.n_axes == 6
        assert chart.features == tuple(['x']*chart.n_axes)
        assert chart.targets == tuple(['y']*chart.n_axes)
        assert chart.hues == tuple(['category']*chart.n_axes)
    
    def test_normalize_to_tuple(self, sample_data: pd.DataFrame) -> None:
        chart = JointChart(
            sample_data, feature='x', target='y', hue='category',
            ncols=2, nrows=3)
        assert chart.n_axes == 6
        assert chart.normalize_to_tuple('x') == ('x', 'x', 'x', 'x', 'x', 'x')

        with pytest.raises(AssertionError) as err:
            chart.normalize_to_tuple(('x', 'x'))
            assert 'not enough values' in str(err)

    def test_itercharts(self, sample_data: pd.DataFrame) -> None:
        chart = JointChart(
            sample_data, feature='x', target='y', hue='category',
            ncols=2, nrows=3)
        axes = chart.axes.flat
        for i, _chart in enumerate(chart.itercharts()):
            assert _chart == chart.charts[i]
            assert chart.axes.ax == axes[i]
    
    def test_specification_limits_iterator(self, sample_data: pd.DataFrame) -> None:
        chart = JointChart(
            sample_data, feature='x', target='y', hue='category',
            ncols=2, nrows=3)
        
        spec_limits = SpecLimits(1.0, 2.0)
        limits = chart.specification_limits_iterator(spec_limits)
        assert isinstance(limits, Generator)
        for limit in limits:
            assert limit == spec_limits
        
        spec_limits = (
            SpecLimits(1.0, 2.0), SpecLimits(3.0, 4.0), SpecLimits(5.0, 6.0), 
            SpecLimits(7.0, 8.0), SpecLimits(9.0, 10.0), SpecLimits(11.0, 12.0))
        limits = chart.specification_limits_iterator(spec_limits)
        for i, limit in enumerate(limits):
            assert limit == spec_limits[i]

        spec_limits = (SpecLimits(1.0, 2.0), SpecLimits(3.0, 4.0))  # Less limits than n_axes
        with pytest.raises(AssertionError) as err:
            limits = chart.specification_limits_iterator(spec_limits)
            next(limits)

    def test_share_axis(self, sample_data: pd.DataFrame) -> None:
        chart = JointChart(
            sample_data, feature='x', target='y', hue='category',
            ncols=2, nrows=3, target_on_y=True, sharex=True, sharey='all')
        assert chart.axes_share_feature == True
        assert chart.axes_share_feature == chart.axes.sharex
        assert chart.axes_share_target == 'all'
        assert chart.axes_share_target == chart.axes.sharey

        chart = JointChart(
            sample_data, feature='x', target='y', hue='category',
            ncols=2, nrows=3, target_on_y=False, sharex='row', sharey='none')
        assert chart.axes_share_feature == 'none'
        assert chart.axes_share_feature == chart.axes.sharey
        assert chart.axes_share_target == 'row'
        assert chart.axes_share_target == chart.axes.sharex

        with pytest.warns(UserWarning):
            chart = JointChart(
                sample_data, feature='x', target='y', hue='category',
                ncols=2, nrows=1, target_on_y=(True, False), sharex=True)

    def test_single_label_allowed(self, sample_data: pd.DataFrame) -> None:
        chart = JointChart(
            sample_data, feature='x', target='y', hue='category',
            ncols=2, nrows=3, sharex='none', sharey='none')
        assert chart.single_label_allowed(is_target=False)
        assert chart.single_label_allowed(is_target=True)

        chart = JointChart(
            sample_data, feature=('x', 'category'), target=('y', 'z'), 
            ncols=2, nrows=1, sharex=True, sharey='none')
        assert chart.single_label_allowed(is_target=False)
        assert not chart.single_label_allowed(is_target=True)

        chart = JointChart(
            sample_data, feature=('x', 'category'), target=('y', 'z'), 
            ncols=2, nrows=1, sharex=True, sharey='none', target_on_y=False)
        assert not chart.single_label_allowed(is_target=False)
        assert chart.single_label_allowed(is_target=True)


class TestChartFormatting:
    """Test the new formatting features for SingleChart and MultivariateChart."""

    @pytest.fixture
    def sample_data(self) -> pd.DataFrame:
        """Create sample data for testing."""
        np.random.seed(42)
        return pd.DataFrame({
            'x': np.random.normal(0, 1, 30),
            'y': np.random.normal(0, 1, 30),
            'category': np.random.choice(['A', 'B', 'C'], 30),
            'facility': np.random.choice(['North', 'South'], 30)
        })

    def test_single_chart_parameter_mapping_standard(self, sample_data: pd.DataFrame) -> None:
        """Test parameter mapping for SingleChart with target_on_y=True."""
        chart = SingleChart(
            source=sample_data,
            target='y',
            feature='x',
            target_on_y=True
        )
        
        # Test standard orientation mapping
        result = chart._map_axis_parameters(
            feature_formatter=lambda x: f"{x:.1f}",
            target_formatter=lambda x: f"{x:.2f}",
            feature_angle=30,
            target_angle=45,
            feature_align='left',
            target_align='right'
        )
        
        xlabel_formatter, ylabel_formatter, xlabel_angle, ylabel_angle, xlabel_align, ylabel_align = result
        
        # With target_on_y=True: feature->x, target->y
        assert xlabel_formatter is not None  # feature formatter goes to x
        assert ylabel_formatter is not None  # target formatter goes to y
        assert xlabel_angle == 30  # feature angle goes to x
        assert ylabel_angle == 45  # target angle goes to y
        assert xlabel_align == 'left'  # feature align goes to x (unchanged)
        assert ylabel_align == 'top'  # target align 'right' maps to 'top' for y-axis

    def test_single_chart_parameter_mapping_swapped(self, sample_data: pd.DataFrame) -> None:
        """Test parameter mapping for SingleChart with target_on_y=False."""
        chart = SingleChart(
            source=sample_data,
            target='y',
            feature='x',
            target_on_y=False
        )
        
        # Test swapped orientation mapping
        result = chart._map_axis_parameters(
            feature_formatter=lambda x: f"{x:.1f}",
            target_formatter=lambda x: f"{x:.2f}",
            feature_angle=30,
            target_angle=45,
            feature_align='left',
            target_align='right'
        )
        
        xlabel_formatter, ylabel_formatter, xlabel_angle, ylabel_angle, xlabel_align, ylabel_align = result
        
        # With target_on_y=False: target->x, feature->y
        assert xlabel_formatter is not None  # target formatter goes to x
        assert ylabel_formatter is not None  # feature formatter goes to y
        assert xlabel_angle == 45  # target angle goes to x
        assert ylabel_angle == 30  # feature angle goes to y
        assert xlabel_align == 'right'  # target align goes to x (unchanged)
        assert ylabel_align == 'bottom'  # feature align 'left' maps to 'bottom' for y-axis

    def test_single_chart_alignment_mapping(self, sample_data: pd.DataFrame) -> None:
        """Test alignment mapping from general to axis-specific alignments."""
        chart = SingleChart(
            source=sample_data,
            target='y',
            feature='x'
        )
        
        # Test all alignment combinations
        test_cases = [
            ('left', 'center', 'left', 'center'),
            ('center', 'left', 'center', 'bottom'),
            ('right', 'right', 'right', 'top')
        ]
        
        for feature_align, target_align, expected_x, expected_y in test_cases:
            result = chart._map_axis_parameters(
                feature_align=feature_align,  # type: ignore
                target_align=target_align  # type: ignore
            )
            
            _, _, _, _, xlabel_align, ylabel_align = result
            assert xlabel_align == expected_x
            assert ylabel_align == expected_y

    def test_single_chart_formatter_integration(self, sample_data: pd.DataFrame) -> None:
        """Test that formatters are properly integrated into the chart."""
        def custom_formatter(x):
            return f"Value: {x:.1f}"
        
        chart = SingleChart(
            source=sample_data,
            target='y',
            feature='x'
        ).plot(
            Scatter,
            alpha=0.5
        ).label(
            target_formatter=custom_formatter,
            feature_formatter=lambda x: f"X: {x:.2f}"
        )
        
        # Verify that the chart was created successfully
        assert hasattr(chart, 'label_facets')
        assert chart.label_facets is not None

    def test_multivariate_chart_parameter_integration(self, sample_data: pd.DataFrame) -> None:
        """Test that MultivariateChart properly integrates formatting parameters."""
        def temp_formatter(x):
            return f"{x:.1f}°C"
        
        chart = MultivariateChart(
            source=sample_data,
            target='y',
            feature='x',
            col='facility'
        ).plot(
            Scatter,
            alpha=0.5
        ).label(
            target_formatter=temp_formatter,
            feature_angle=25,
            target_align='right'
        )
        
        # Verify that the chart was created successfully
        assert hasattr(chart, 'label_facets')
        assert chart.label_facets is not None

    def test_chart_with_all_formatting_options(self, sample_data: pd.DataFrame) -> None:
        """Test chart with all formatting options combined."""
        chart = SingleChart(
            source=sample_data,
            target='y',
            feature='category',
            categorical_feature=True,
            hue='facility'
        ).plot(
            Scatter,
            alpha=0.7
        ).label(
            fig_title='Test Chart with All Options',
            target_label='Target Variable',
            feature_label='Feature Variable',
            target_formatter=lambda x: f"{x:.2f} units",
            feature_formatter=lambda x: f"Cat {x}",
            target_angle=15,
            feature_angle=45,
            target_align='center',
            feature_align='right'
        )
        
        # Verify chart creation and label facets
        assert hasattr(chart, 'label_facets')
        assert chart.label_facets.xlabel_angle == 45
        assert chart.label_facets.ylabel_angle == 15



"""
![pypi](https://badge.fury.io/py/daspi.svg)
![licence](https://img.shields.io/github/license/j4ggr/daspi.svg)
![GitHub last commit](https://img.shields.io/github/last-commit/j4ggr/daspi)
![downloads](https://img.shields.io/pypi/dm/daspi)

![logo](https://raw.githubusercontent.com/j4ggr/DaSPi/main/docs/img/logo.svg)

# Data analysis, Statistics and Process improvements (DaSPi)

Visualize and analyze your data with DaSPi. This package is designed for users who want to find relevant influencing factors in processes and validate improvements.
This package offers many [Six Sigma](https://en.wikipedia.org/wiki/Six_Sigma) tools based on the following packages:

- [pandas](https://pandas.pydata.org/)
- [numpy](https://numpy.org/)
- [matplotlib](https://matplotlib.org/)
- [scipy](https://scipy.org/)
- [statsmodels](https://www.statsmodels.org/stable/index.html)

The goal of this package is to be easy to use and flexible so that it can be adapted to a wide array of data analysis tasks.

## Why DaSPi?

There are great packages for data analysis and visualization in Python, such as [Pandas](https://pandas.pydata.org/pandas-docs/stable), [Seaborn](https://seaborn.pydata.org/index.html), [Altair](https://altair-viz.github.io/), [Statsmodels](https://www.statsmodels.org/stable/), [Scipy](https://docs.scipy.org/doc/scipy/), [Pinguins](https://pingouin-stats.org/). But most of the time they work not directly with each other. Wouldn't it be great if you could use all of these packages together in one place? That's where DaSPi comes in. DaSPi is a Python package that provides a unified interface for data analysis, statistics and visualization. It allows you to use all of the great packages mentioned above together in one place, making it easier to explore and understand your data.

## Features

- **Ease of Use:** DaSPi is designed to be easy to use, even for beginners. It provides a simple and intuitive interface that makes it easy to get started with data analysis.
- **Visualization:** DaSPi provides a wide range of visualization options, including multivariate charts, joint charts, and useful precast. This makes it easy to explore and understand your data in a visual way.
- **Statistics:** DaSPi provides a wide range of statistical functions and tests, including hypothesis testing, confidence intervals, and regression analysis. This makes it easy to explore and understand your data in a statistical way.
- **Open Source:** DaSPi is open source, which means that it is free to use and modify. This makes it a great option for users who want to customize the package to their specific needs.

This Package contains following submodules:

- **plotlib:** Visualizations with Matplotlib, where the division by color, marker size or shape as well as rows and columns subplots are automated depending on the given categorical data. Any plots can also be combined, such as scatter with contour plot, violin with error bars or other creative combinations.
- **anova:** analysis of variance (ANOVA), which is used to compare the variance within and between of two or more groups, or the effects of different treatments on a response variable. It also includes a function for calculating the variance inflation factor (VIF) for linear regression models. The main class is LinearModel, which provides methods for fitting linear regression with interactions and automatically elimiinating insignificant variables.
- **statistics:** applied statistics, hypothesis test, confidence calculations and monte-carlo simulation. It also includes estimation for process capability and capability index.
- **datasets:** data for exersices. It includes different datasets that can be used for testing and experimentation.

## Usage

### Visualization

To use DaSPi, you can import the package and start exploring your data. Here is an example of how to use DaSPi to visualize a dataset:

```python
import daspi as dsp
df = dsp.load_dataset('iris')

chart = dsp.MultivariateChart(
        source=df,
        target='length',
        feature='width',
        hue='species',
        col='leaf',
        markers=('x',)
    ).plot(
        dsp.GaussianKDEContour
    ).plot(
        dsp.Scatter
    ).label(
        feature_label='leaf width (cm)',
        target_label='leaf length (cm)',
    )
```

![Iris sepal length species](https://raw.githubusercontent.com/j4ggr/DaSPi/main/docs/img/iris_contour_size-leaf-species.png)

### ANOVA

Do some ANOVA and statistics on a dataset. Run the example below in a Jupyther Notebook to see the results.

```python
df = dsp.load_dataset('painkillers-dissolution')
model = dsp.LinearModel(
    source=df,
    target='dissolution',
    features=['employee', 'stirrer', 'brand', 'catalyst', 'water'],
    covariates=['temperature', 'preparation'],
    order=2)
df_gof = pd.concat(model.recursive_elimination())

dsp.ResidualsCharts(model).plot().stripes().label(info=True)
dsp.ParameterRelevanceCharts(model).plot().label(info=True)
model
```

**Formula:**

dissolution ~ 16.0792 + 2.3750 employee[T.B] + 0.8375 employee[T.C] + 10.7500 brand[T.OuchAway] - 3.8000 water[T.tap] - 5.7167 brand[T.OuchAway]:water[T.tap]

**Model Summary**

| Hierarchical | Least Parameter |  P Least |        S |        AIC |       R² |   R² Adj |  R² Pred |
| -----------: | --------------: | -------: | -------: | ---------: | -------: | -------: | -------: |
|         True |        employee | 0.023298 | 2.374693 | 224.835935 | 0.857379 | 0.840400 | 0.813719 |

**Parameter Statistics**

|                               |      Coef |  Std Err |         T |        P |    CI Low |    CI Upp |
| ----------------------------: | --------: | -------: | --------: | -------: | --------: | --------: |
|                     Intercept | 16.079167 | 0.839581 | 19.151424 | 0.000000 | 14.384824 | 17.773509 |
|                 employee[T.B] |  2.375000 | 0.839581 |  2.828793 | 0.007133 |  0.680657 |  4.069343 |
|                 employee[T.C] |  0.837500 | 0.839581 |  0.997522 | 0.324224 | -0.856843 |  2.531843 |
|              brand[T.OuchAway] | 10.750000 | 0.969464 | 11.088598 | 0.000000 |  8.793542 | 12.706458 |
|                  water[T.tap] | -3.800000 | 0.969464 | -3.919690 | 0.000321 | -5.756458 | -1.843542 |
| brand[T.OuchAway]:water[T.tap] | -5.716667 | 1.371030 | -4.169616 | 0.000149 | -8.483516 | -2.949817 |

**Analysis of Variance**

|      Source |   DF |         SS |         MS |          F |        P |       n² |
| ----------: | ---: | ---------: | ---------: | ---------: | -------: | -------: |
|    employee |    2 |  46.431667 |  23.215833 |   4.116891 | 0.023298 | 0.027960 |
|       brand |    1 | 747.340833 | 747.340833 | 132.526821 | 0.000000 | 0.450027 |
|       water |    1 | 532.000833 | 532.000833 |  94.340328 | 0.000000 | 0.320355 |
| brand:water |    1 |  98.040833 |  98.040833 |  17.385695 | 0.000149 | 0.059037 |
|    Residual |   42 | 236.845000 |   5.639167 |        nan |      nan | 0.142621 |

**Variance Inflation Factor**

|             |   DF |      VIF |     GVIF | Threshold | Collinear |              Method |
| ----------: | ---: | -------: | -------: | --------: | --------: | ------------------: |
|   Intercept |    1 | 5.000000 | 2.236068 |  2.236068 |      True |           R_squared |
|    employee |    2 | 1.000000 | 1.000000 |  1.495349 |     False |         generalized |
|       brand |    1 | 1.000000 | 1.000000 |  2.236068 |     False |           R_squared |
|       water |    1 | 1.000000 | 1.000000 |  2.236068 |     False |           R_squared |
| brand:water |    1 | 1.000000 | 1.000000 |  2.236068 |     False | single_order-2_term |

![ANOVA dissolution residuals](https://raw.githubusercontent.com/j4ggr/DaSPi/main/docs/img/anova_dissolution_residues.png)

![ANOVA dissolution parameters](https://raw.githubusercontent.com/j4ggr/DaSPi/main/docs/img/anova_dissolution_params.png)

### Process capability

Analyze process variation and other key performance indicators for process capacity.

```python
df = dsp.load_dataset('drop_card')
spec_limits = dsp.SpecLimits(0, float(df.loc[0, 'usl']))
target = 'distance'

chart = dsp.ProcessCapabilityAnalysisCharts(
        source=df,
        target=target,
        spec_limits=spec_limits,
        hue='method'
    ).plot(
    ).stripes(
    ).label(
        fig_title='Process Capability Analysis',
        sub_title='Drop Card Experiment',
        target_label='Distance (cm)',
        info=True
    )

samples_parallel = df[df['method']=='parallel'][target]
samples_series = df[df['method']=='perpendicular'][target]
pd.concat([
    dsp.ProcessEstimator(samples_parallel, spec_limits).describe(),
    dsp.ProcessEstimator(samples_series, spec_limits).describe()],
    axis=1,
    ignore_index=True,
).rename(
    columns={0: 'parallel', 1: 'perpendicular'}
)
```

|           |   parallel | perpendicular |
| --------: | ---------: | ------------: |
| n_samples |         20 |            20 |
| n_missing |          0 |             0 |
|      n_ok |         18 |            20 |
|     n_nok |          2 |             0 |
|  n_errors |          0 |             0 |
|        ok |    90.00 % |      100.00 % |
|       nok |    10.00 % |        0.00 % |
|  nok_norm |     8.01 % |        3.73 % |
|   nok_fit |     7.24 % |        5.77 % |
|       min |        8.5 |          17.5 |
|       max |       83.0 |          73.0 |
|      mean |     42.935 |        48.485 |
|    median |      40.75 |          52.5 |
|       std |  22.666583 |     17.359489 |
|       sem |   5.068402 |        3.8817 |
|    excess |  -0.900801 |     -1.236078 |
|  p_excess |   0.288757 |      0.072573 |
|      skew |    0.19252 |     -0.377538 |
|    p_skew |   0.690373 |      0.438723 |
|      p_ad |   0.754044 |      0.098371 |
|      dist |    lognorm |      logistic |
|    p_ks |   0.964797 |      0.744326 |
|  strategy |       norm |          norm |
|       lcl | -25.064748 |     -3.593468 |
|       ucl | 110.934748 |    100.563468 |
|       lsl |          0 |             0 |
|       usl |       80.0 |          80.0 |
|        cp |   0.588237 |      0.768072 |
|       cpk |   0.545076 |      0.605145 |
|         Z |   1.635227 |      1.815434 |
|      Z_lt |   0.135227 |      0.315434 |

![Process Capability Analysis](https://raw.githubusercontent.com/j4ggr/DaSPi/main/docs/img/cpk-analysis_drop-card.png)

## About DaSPi

DaSPi was created and is actively maintained by **Reto Jäggli**, a Data Scientist at Festo Microtechnology AG. 
Much of the development happens during spare time, driven by a passion for making data analysis, statistics, and process improvement more accessible and integrated.

Contributions to DaSPi are very welcome!
If you find bugs or have ideas for improvements, please report them or submit pull requests on the [GitHub repository](https://github.com/j4ggr/DaSPi), where the full source code is also available for review.

**Important Notice:**  
DaSPi is still under heavy development and may contain hidden bugs. 
While every effort is made to ensure reliability, no warranty is provided. 
The results obtained using DaSPi should be double-checked with other trusted statistical software whenever possible. 
Where applicable, DaSPi acts as a convenient wrapper around well-established packages such as pandas, numpy, matplotlib, scipy, and statsmodels, leveraging their robustness and functionality.
"""

from ._version import __version__

from .strings import STR


from .constants import KW
from .constants import RE
from .constants import DIST
from .constants import LINE
from .constants import COLOR
from .constants import LABEL
from .constants import ANOVA
from .constants import PLOTTER
from .constants import DEFAULT
from .constants import CATEGORY
from .constants import SIGMA_DIFFERENCE


from .doe import Factor
from .doe import FullFactorialDesignBuilder
from .doe import FullFactorial2kDesignBuilder
from .doe import FractionalFactorialDesignBuilder
from .doe import get_default_generators


from .statistics.montecarlo import Binning
from .statistics.montecarlo import SpecLimits
from .statistics.montecarlo import Specification
from .statistics.montecarlo import round_to_nearest
from .statistics.montecarlo import RandomProcessValue
from .statistics.montecarlo import inclination_displacement

from .statistics.confidence import sem
from .statistics.confidence import cp_ci
from .statistics.confidence import cpk_ci
from .statistics.confidence import fit_ci
from .statistics.confidence import mean_ci
from .statistics.confidence import stdev_ci
from .statistics.confidence import median_ci
from .statistics.confidence import variance_ci
from .statistics.confidence import proportion_ci
from .statistics.confidence import bonferroni_ci
from .statistics.confidence import delta_mean_ci
from .statistics.confidence import prediction_ci
from .statistics.confidence import delta_stdev_ci
from .statistics.confidence import delta_variance_ci
from .statistics.confidence import confidence_to_alpha
from .statistics.confidence import delta_proportions_ci

from .statistics.hypothesis import f_test
from .statistics.hypothesis import t_test
from .statistics.hypothesis import skew_test
from .statistics.hypothesis import all_normal
from .statistics.hypothesis import levene_test
from .statistics.hypothesis import position_test
from .statistics.hypothesis import variance_test
from .statistics.hypothesis import kurtosis_test
from .statistics.hypothesis import proportions_test
from .statistics.hypothesis import mean_stability_test
from .statistics.hypothesis import anderson_darling_test
from .statistics.hypothesis import kolmogorov_smirnov_test
from .statistics.hypothesis import variance_stability_test

from .statistics.estimation import Loess
from .statistics.estimation import Lowess
from .statistics.estimation import GageEstimator
from .statistics.estimation import ProcessEstimator
from .statistics.estimation import root_sum_squares
from .statistics.estimation import estimate_resolution
from .statistics.estimation import DistributionEstimator
from .statistics.estimation import estimate_distribution
from .statistics.estimation import MeasurementUncertainty
from .statistics.estimation import estimate_kernel_density
from .statistics.estimation import estimate_kernel_density_2d
from .statistics.estimation import LocationDispersionEstimator
from .statistics.estimation import estimate_capability_confidence


from .plotlib import style

from .plotlib.classify import Dodger
from .plotlib.classify import HueLabel
from .plotlib.classify import SizeLabel
from .plotlib.classify import ShapeLabel

from .plotlib.plotter import Box
from .plotlib.plotter import Bar
from .plotlib.plotter import Line
from .plotlib.plotter import Stem
from .plotlib.plotter import Pareto
from .plotlib.plotter import Jitter
from .plotlib.plotter import Plotter
from .plotlib.plotter import Scatter
from .plotlib.plotter import Violine
from .plotlib.plotter import Beeswarm
from .plotlib.plotter import Errorbar
from .plotlib.plotter import MeanTest
from .plotlib.plotter import LoessLine
from .plotlib.plotter import StripeLine
from .plotlib.plotter import StripeSpan
from .plotlib.plotter import HideSubplot
from .plotlib.plotter import SkipSubplot
from .plotlib.plotter import SpreadWidth
from .plotlib.plotter import Probability
from .plotlib.plotter import BlandAltman
from .plotlib.plotter import GaussianKDE
from .plotlib.plotter import QuantileBoxes
from .plotlib.plotter import VariationTest
from .plotlib.plotter import ProportionTest
from .plotlib.plotter import CenterLocation
from .plotlib.plotter import TransformPlotter
from .plotlib.plotter import StandardErrorMean
from .plotlib.plotter import ConfidenceInterval
from .plotlib.plotter import ParallelCoordinate
from .plotlib.plotter import GaussianKDEContour
from .plotlib.plotter import LinearRegressionLine
from .plotlib.plotter import CategoricalObservation
from .plotlib.plotter import CapabilityConfidenceInterval

from .plotlib.facets import AxesFacets
from .plotlib.facets import flat_unique
from .plotlib.facets import LabelFacets
from .plotlib.facets import StripesFacets

from .plotlib.chart import Chart
from .plotlib.chart import JointChart
from .plotlib.chart import SingleChart
from .plotlib.chart import MultivariateChart

from .plotlib.precast import GageRnRCharts
from .plotlib.precast import GageStudyCharts
from .plotlib.precast import ResidualsCharts
from .plotlib.precast import PairComparisonCharts
from .plotlib.precast import PairwiseMatrixCharts
from .plotlib.precast import ParameterRelevanceCharts
from .plotlib.precast import BivariateUnivariateCharts
from .plotlib.precast import ProcessCapabilityAnalysisCharts

from .anova.convert import get_term_name
from .anova.convert import frames_to_html

from .anova.tables import uniques
from .anova.tables import anova_table
from .anova.tables import terms_effect
from .anova.tables import terms_probability
from .anova.tables import variance_inflation_factor

from .anova.model import LinearModel
from .anova.model import GageRnRModel
from .anova.model import hierarchical
from .anova.model import GageStudyModel
from .anova.model import is_main_parameter

from .datasets import load_dataset
from .datasets import list_dataset






import matplotlib as mpl
import matplotlib.pyplot as plt

from typing import List
from pathlib import Path

from .appearance import *

from .classify import *

from .plotter import *

from .facets import *

from .chart import *

from .precast import *

# TODO: some alpha fails for vspan when using daspi and ggplot2 styles
style.use('daspi')

